from numpy import array
from numpy import zeros
from numpy import vstack
from numpy import hstack
from numpy import identity
from numpy import asarray
from numpy.linalg import pinv
from numpy.linalg import svd

from scipy.sparse import csr_matrix
from scipy.sparse import diags
from scipy.sparse import vstack as svstack

from compas.numerical import connectivity_matrix

from compas.utilities import reverse_geometric_key

from compas.geometry import Point
from compas.geometry import distance_point_point_xy

from compas_tno.algorithms import check_independents
from compas_tno.algorithms import check_horizontal_loads
from compas_tno.algorithms import find_independents

from compas_tno.utilities import apply_radial_symmetry
from compas_tno.utilities import apply_symmetry_from_axis
from compas_tno.utilities import find_sym_axis_in_rect_patterns
from compas_tno.utilities import build_symmetry_transformation

import time


class Problem():
    """
    The ``Problem`` class stores the matrices used in the optimisation. These are listed as parameters of the class and described below.

    Parameters
    ----------

    None

    Attributes
    ----------

    q : array(m x 1)
        The vector of force densities
    m : int
        The number of edges
    n : int
        The number of vertices
    nb : int
        The number of fixed vertices
    E : array(2n x m)
        The horizontal equilibrium matrix
    C : array(n x m)
        The connectivity matrix
    Ct : array(m x n)
        The transposed connectivity matrix
    Ci : array(m x ni)
        The sliced connectivity matrix with regards to the internal vertices
    Cit : array(ni x m)
        The transpose of Ci
    Cb : array(nb x m)
        The sliced connectivity matrix with regards to the constrained (support) vertices
    U : array(m x m)
        The diagonal matrix of coorrdinate differences in the x-direction
    V : array(m x m)
        The diagonal matrix of coorrdinate differences in the y-direction
    P : array(n x 3)
        The applied external loads in all nodes
    free : list
        The list with the index of the free vertices
    fixed : list
        The list with the index of the fixed vertices
    phfree : array(ni x 2)
        The applied horizonal loads on the free vertices
    ph : array(ni x 2)
        A column vector with the the applied horizonal loads on the free vertices
    lb : array(n x 1)
        The lower-bound (intrados) limit for the nodes
    ub : array(n x 1)
        The upper-bound (extrados) limit for the nodes
    lb0 : array(n x 1)
        The lower-bound (intrados) limit for the nodes at the starting point (can be modified if thickness in the objective)
    ub0 : array(n x 1)
        The upper-bound (extrados) limit for the nodes at the starting point (can be modified if thickness in the objective)
    s : array(n x 1)
        The middle surface of the nodes
    X : array(n x 3)
        The nodal position of the vertices of the network
    x0 : array(n x 1)
        The x-position of the vertices in the network
    y0 : array(n x 1)
        The y-position of the vertices in the network
    free_x : list
        index of the vertices free to move in x
    free_y : list
        index of the vertices free to move in y
    rol_x : list
        index of the vertices constrained partially on x.
    rol_y : list
        index of the vertices constrained partially on y.
    Citx : array
        Connectivity matrix sliced transposed on the vertices free on x.
    City : array
        Connectivity matrix sliced transposed on the vertices free on y.
    Cbtx : array
        Connectivity matrix sliced transposed on the vertices partially fixed on x.
    Cbty : array
        Connectivity matrix sliced transposed on the vertices partially fixed on y.
    xlimits : array(n x 1)
        Limits on the x-direction in which the nodes can move
    ylimits : array(n x 1)
        Limits on the y-direction in which the nodes can move
    qmin : array(m x 1)
        Lower-bounds of the force densities in the edges
    qmax : array(m x 1)
        Upper-bounds of the force densities in the edges
    k_i : dict
        key-index dictionary
    uv_i : dict
        uv-index dictionary
    i_uv : dict
        index-uv dictionary
    ind : list
        List with the index of the independent edges
    k : int
        Number of independents in the problem
    dep : list
        List with the index of the dependent edges
    B : array(m x k)
        Matrix transforming the force densities in the independent edges to all force densities

    """

    def __init__(self):

        self.q = None
        self.m = None
        self.n = None
        self.ni = None
        self.nb = None
        self.E = None
        self.C = None
        self.Ct = None
        self.Ci = None
        self.Cit = None
        self.Cb = None
        self.U = None
        self.V = None
        self.P = None
        self.free = None
        self.fixed = None
        self.ph = None
        self.lb = None
        self.ub = None
        self.lb0 = None
        self.ub0 = None
        self.s = None
        self.X = None
        self.x0 = None
        self.y0 = None
        self.free_x = None
        self.free_y = None
        self.rol_x = None
        self.rol_y = None
        self.Citx = None
        self.City = None
        self.Cbtx = None
        self.Cbty = None
        self.xlimits = None
        self.ylimits = None
        self.qmin = None
        self.qmax = None
        self.k_i = None
        self.uv_i = None
        self.i_uv = None
        self.ind = None
        self.k = None
        self.dep = None
        self.B = None

        pass

    @classmethod
    def from_formdiagram(cls, form, indset=None, printout=False, find_inds=False):
        """Initialise the problem object form a form diagram."""
        problem = cls()

        return problem


def initialise_form(form, find_inds=True, method='SVD', printout=False, tol=None):
    """ Initialise the problem for a Form-Diagram and return the FormDiagram with independent edges assigned and the matrices relevant to the equilibrium problem.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram
    find_inds : bool, optional
        Whether or not independents should be found (fixed diagram), by default True
    method : str, optional
        Method to find independent edges, the default is 'SVD'. More options to come.
    printout : bool, optional
        Whether or not prints should appear on the screen, by default False
    tool : float, optional
        Tolerance od independents, by default None

    Returns
    -------
    M : :class:`~compas_tno.problems.Problem`
        The object with all the matrices essential to the analysis.

    Note
    -------
    The FormDiagram is updated in place. Check ``initialise_problem_general`` for more info.

    """

    M = initialise_problem_general(form)

    if find_inds:
        adapt_problem_to_fixed_diagram(M, form, method=method, printout=printout, tol=tol)

    return M


def initialise_problem_general(form):
    """ Initialise the problem for a given Form-Diagram building the main matrices used in the subsequent analysis.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram.

    Returns
    -------
    args
        List of matrices and vectors used to perform the optimisation.

    """

    # Mapping

    k_i = form.key_index()
    uv_i = form.uv_index()
    i_uv = form.index_uv()

    # Vertices and edges

    n = form.number_of_vertices()
    m = len(list(form.edges_where({'_is_edge': True})))
    fixed = [k_i[key] for key in form.fixed()]
    nb = len(fixed)
    edges = [(k_i[u], k_i[v]) for u, v in form.edges_where({'_is_edge': True})]
    free = list(set(range(n)) - set(fixed))
    ni = len(free)

    q = array([form.edge_attribute((u, v), 'q') for u, v in form.edges_where({'_is_edge': True})]).reshape(-1, 1)  # review need of 'is_edge': True
    qmax = array([form.edge_attribute((u, v), 'qmax') for u, v in form.edges_where({'_is_edge': True})]).reshape(-1, 1)
    qmin = array([form.edge_attribute((u, v), 'qmin') for u, v in form.edges_where({'_is_edge': True})]).reshape(-1, 1)

    # Co-ordinates, loads and constraints

    xyz = zeros((n, 3))
    x = zeros((n, 1))
    y = zeros((n, 1))
    z = zeros((n, 1))
    s = zeros((n, 1))
    px = zeros((n, 1))
    py = zeros((n, 1))
    pz = zeros((n, 1))
    lb = zeros((n, 1))
    ub = zeros((n, 1))
    xlimits = zeros((n, 2))
    ylimits = zeros((n, 2))

    for key, vertex in form.vertex.items():
        i = k_i[key]
        xyz[i, :] = form.vertex_coordinates(key)
        x[i] = vertex.get('x')
        y[i] = vertex.get('y')
        z[i] = vertex.get('z')
        px[i] = vertex.get('px', 0)
        py[i] = vertex.get('py', 0)
        pz[i] = vertex.get('pz', 0)
        s[i] = vertex.get('target', 0)
        if abs(s[i]) < 1e-6:
            s[i] = 0.0
        xlimits[i, 0] = vertex.get('xmin', None)
        xlimits[i, 1] = vertex.get('xmax', None)
        ylimits[i, 0] = vertex.get('ymin', None)
        ylimits[i, 1] = vertex.get('ymax', None)
        lb[i] = vertex.get('lb', None)
        ub[i] = vertex.get('ub', None)

    # Partial supports, or rollers.

    rol_x = []
    rol_y = []

    for key in form.vertices_where({'rol_x': True}):
        rol_x.append(k_i[key])
    for key in form.vertices_where({'rol_y': True}):
        rol_y.append(k_i[key])

    free_x = list(set(free) - set(rol_x))
    free_y = list(set(free) - set(rol_y))

    # C and E matrices

    C = connectivity_matrix(edges, 'csr')
    Ci = C[:, free]
    Cb = C[:, fixed]
    Cbtx = C[:, rol_x].transpose()
    Cbty = C[:, rol_y].transpose()
    Ct = C.transpose()
    Cit = Ci.transpose()
    Citx = Ct[free_x, :]
    City = Ct[free_y, :]
    uvw = C.dot(xyz)
    U = diags(uvw[:, 0].flatten())
    V = diags(uvw[:, 1].flatten())
    E = svstack((Citx.dot(U), City.dot(V))).toarray()

    # Settings of a free to move problem 'all q are variable'

    ind = list(range(m))
    dep = []
    k = m
    B = identity(m)
    d = zeros((m, 1))

    # Permutation matrix to fixed and free keys
    # with this matrix z = [Pmatrix]z will have the first ni nodes as free
    # and the last nb nodes as fixed

    Pfree = zeros((n, ni))
    Pfixed = zeros((n, nb))
    Pfree[free, :] = identity(ni)
    Pfixed[fixed, :] = identity(nb)
    Pmatrix = hstack([Pfree, Pfixed])

    # Create Class and build matrices

    problem = Problem()

    problem.q = q
    problem.m = m
    problem.n = n
    problem.ni = ni
    problem.nb = nb
    problem.E = E
    problem.C = C
    problem.Ct = Ct
    problem.Ci = Ci
    problem.Cit = Cit
    problem.Cb = Cb
    problem.U = U
    problem.V = V
    problem.P = hstack([px, py, pz])
    problem.free = free
    problem.fixed = fixed
    problem.ph = vstack([px[free_x], py[free_y]])
    problem.lb = lb
    problem.ub = ub
    problem.lb0 = lb
    problem.ub0 = ub
    problem.s = s
    problem.X = hstack([x, y, z])
    problem.x0 = x
    problem.y0 = y
    problem.free_x = free_x
    problem.free_y = free_y
    problem.rol_x = rol_x
    problem.rol_y = rol_y
    problem.Citx = Citx
    problem.City = City
    problem.Cbtx = Cbtx
    problem.Cbty = Cbty
    problem.xlimits = xlimits
    problem.ylimits = ylimits
    problem.qmin = qmin
    problem.qmax = qmax
    problem.k_i = k_i
    problem.uv_i = uv_i
    problem.i_uv = i_uv
    problem.ind = ind
    problem.k = k
    problem.dep = dep
    problem.B = B
    problem.d = d
    problem.Pmatrix = Pmatrix
    # problem.Bfixed = Bfixed

    return problem


def adapt_problem_to_fixed_diagram(problem, form, method='SVD', printout=False, tol=None):
    """Adapt the problem assuming that the form diagram is fixed in plan.

    Parameters
    ----------
    problem : :class:`~compas_tno.problems.Problem`
        Matrices of the problem
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to be analysed
    method : str, optional
        Method to find independent edges, the default is 'SVD'. More options to come.
    printout : bool, optional
        If prints should show in the screen, by default False
    tol : float, optional
        Tolerance of the singular values, by default None

    """

    ind = []
    tol_old_ind = 1e-3

    start_time = time.time()

    # Independent and dependent branches

    if form.attributes['indset']:
        # check if it is a string and "restaure the points"
        indset = [a if not isinstance(a, str) else reverse_geometric_key(a) for a in form.attributes['indset']]
        ind = []

        for u, v in form.edges_where({'_is_edge': True}):
            index = problem.uv_i[(u, v)]
            edgemid = Point(*(form.edge_midpoint(u, v)[:2] + [0]))
            for pt in indset:
                if distance_point_point_xy(edgemid, pt) < tol_old_ind:
                    ind.append(index)
                    break
            if index in ind:
                indset.remove(pt)

        if printout:
            print('Loaded {} previous independents'.format(len(form.attributes['indset'])))
            print('Found {} independents in the new pattern'.format(len(ind)))
        if len(form.attributes['indset']) != len(ind):
            print('Did not match problem inds')
            ind = find_independents(problem.E, method=method, tol=tol)
    else:
        ind = find_independents(problem.E, method=method, tol=tol)

    k = len(ind)
    dep = list(set(range(problem.m)) - set(ind))

    elapsed_time = time.time() - start_time

    if printout:
        print('Reduced problem to {0} force variables with ind. edges'.format(k))
        print('Elapsed Time: {0:.1f} sec'.format(elapsed_time))

    points = []
    for u, v in form.edges_where({'_is_edge': True}):
        if problem.uv_i[(u, v)] in ind:
            form.edge_attribute((u, v), 'is_ind', True)
            points.append(Point(*(form.edge_midpoint(u, v)[:2] + [0])))
        else:
            form.edge_attribute((u, v), 'is_ind', False)
    form.attributes['indset'] = points

    rcond = 1e-17
    if tol:
        rcond = tol
    Ed = problem.E[:, dep]
    Edinv = -csr_matrix(pinv(problem.E[:, dep], rcond=rcond))
    Ei = csr_matrix(problem.E[:, ind])
    B = zeros((problem.m, k))
    B[dep] = Edinv.dot(Ei).toarray()
    B[ind] = identity(k)

    d = zeros((problem.m, 1))
    d[dep] = -Edinv.dot(problem.ph)  # q = Bqi + d | d = Ed(-1)*ph

    if any(problem.ph):
        check_hor = check_horizontal_loads(problem.E, problem.ph)
        if check_hor:
            print('Horizontal Loads can be taken!')
        else:
            print('Horizontal Loads are not suitable for this FD!')

    problem.ind = ind
    problem.k = k
    problem.dep = dep
    problem.B = B
    problem.Edinv = Edinv
    problem.Ed = Ed
    problem.Ei = Ei
    problem.d = d
    problem.d0 = d
    problem.time_inds = elapsed_time

    return


def adapt_problem_to_sym_diagram(problem, form, list_axis_symmetry=None, center=None, correct_loads=True, printout=False):
    """Adapt the problem assuming that the form diagram is symmetric.

    Parameters
    ----------
    problem : :class:`~compas_tno.problems.Problem`
        The problem with matrices for calculation
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to analyse
    list_axis_symmetry : [list], optional
        List of the axis of symmetry to consider, by default None
    center : [list], optional
        The center of the pattern if it's a circula pattern, by default None
    correct_loads : bool, optional
        If update should be done in the applied loads regarding the symmetry, by default True
    printout : bool, optional
        If prints should show in the screen, by default False

    """

    start_time = time.time()

    apply_sym_to_form(form, list_axis_symmetry, center, correct_loads)

    Esym = build_symmetry_transformation(form, printout=False)
    mapsym = form.build_symmetry_map()
    ind = sorted(list(mapsym.values()))

    k = len(ind)
    dep = list(set(range(problem.m)) - set(ind))

    elapsed_time = time.time() - start_time

    if printout:
        print('Reduced problem to {0} force variables by SYM'.format(k))
        print('Elapsed Time: {0:.1f} sec'.format(elapsed_time))

    for u, v in form.edges_where({'_is_edge': True}):
        form.edge_attribute((u, v), 'is_ind', True if problem.uv_i[(u, v)] in ind else False)

    B = Esym

    problem.ind = ind
    problem.k = k
    problem.dep = dep
    problem.B = B

    return


def adapt_problem_to_sym_and_fixed_diagram(problem, form, method='SVD', list_axis_symmetry=None, center=None, correct_loads=True, printout=False, tol=None):
    """ Adapt the problem assuming that the form diagram is symmetric and fixed in plane.

    Parameters
    ----------
    problem : :class:`~compas_tno.problems.Problem`
        The problem with matrices for calculation
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to analyse
    method : str, optional
        Method to find independent edges, the default is 'SVD'. More options to come.
    list_axis_symmetry : [list], optional
        List of the axis of symmetry to consider, by default None
    center : [list], optional
        The center of the pattern if it's a circula pattern, by default None
    correct_loads : bool, optional
        If update should be done in the applied loads regarding the symmetry, by default True
    printout : bool, optional
        If prints should show in the screen, by default False

    """

    start_time = time.time()

    adapt_problem_to_fixed_diagram(problem, form, method=method, printout=printout, tol=tol)

    apply_sym_to_form(form, list_axis_symmetry, center, correct_loads)

    ind = problem.ind
    k = problem.k
    i_uv = problem.i_uv
    uv_i = problem.uv_i
    unique_sym = []
    ind_reduc = []

    for index in ind:
        u, v = i_uv[index]
        index_sym = form.edge_attribute((u, v), 'sym_key')
        if index_sym not in unique_sym:
            unique_sym.append(index_sym)
            ind_reduc.append(index)

    k_sym = len(ind_reduc)
    Bsym = zeros((k, k_sym))

    j = 0
    for key in unique_sym:
        edges_sym = [uv_i[(u, v)] for u, v in form.edges_where({'sym_key': key})]
        col = zeros((k, 1))
        for i, ind_i in enumerate(ind):
            if ind_i in edges_sym:
                col[i] = 1.0
        Bsym[:, [j]] = col
        j += 1

    B = problem.B.dot(Bsym)
    ind = ind_reduc
    k = k_sym
    dep = list(set(range(problem.m)) - set(ind))

    elapsed_time = time.time() - start_time

    if printout:
        print('Reduced problem to {0} force variables'.format(k))
        print('Elapsed Time: {0:.1f} sec'.format(elapsed_time))

    for u, v in form.edges_where({'_is_edge': True}):
        form.edge_attribute((u, v), 'is_ind', True if problem.uv_i[(u, v)] in ind else False)

    problem.ind = ind
    problem.k = k
    problem.dep = dep
    problem.B = B

    return


def apply_sym_to_form(form, list_axis_symmetry=None, center=None, correct_loads=True):
    """ Apply symmetry to the form diagram.

    Parameters
    ----------
    problem : :class:`~compas_tno.problems.Problem`
        The problem with matrices for calculation
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to analyse
    list_axis_symmetry : [list], optional
        List of the axis of symmetry to consider, by default None
    center : [list], optional
        The center of the pattern if it's a circula pattern, by default None
    correct_loads : bool, optional
        If update should be done in the applied loads regarding the symmetry, by default True

    """

    if not list_axis_symmetry:
        data = form.parameters
        if data['type'] in ['radial_fd', 'radial_spaced_fd', 'spiral_fd']:
            apply_radial_symmetry(form, center=center, correct_loads=correct_loads)
        elif data['type'] in ['cross_fd', 'fan_fd', 'cross_diagonal', 'cross_with_diagonal', 'ortho_fd']:
            list_axis_symmetry = find_sym_axis_in_rect_patterns(data)
            print('Axis of Symmetry identified:', list_axis_symmetry)
        else:
            print('Symmetry applied to a unknown form diagram type')
            raise NotImplementedError

    if list_axis_symmetry:
        apply_symmetry_from_axis(form, list_axis_symmetry=list_axis_symmetry, correct_loads=correct_loads)

    return


def check_bad_independents(problem, form, printout=False, eps=10e+6):

    keep_inds = []

    for i_ind in range(len(problem.ind)):
        column_i = abs(problem.B[:, i_ind]).flatten()
        max_col = max(column_i)
        if max_col < eps:
            keep_inds.append(i_ind)

    if len(keep_inds) == problem.m:
        return
    else:
        print('original inds:', len(problem.ind), problem.ind)
        print('Updating the relevant independent')
        inds = []
        for i_ind in keep_inds:
            inds.append(problem.ind[i_ind])

        q0 = zeros((problem.m, 1))
        form.edges_attribute('is_ind', False)
        for i, edge in enumerate(form.edges_where({'_is_edge': True})):
            if i in inds:
                form.edge_attribute(edge, 'is_ind', True)
            elif i in problem.ind:
                q0[i] = form.edge_attribute(edge, 'q')

        print(q0)

        problem.ind = inds
        problem.dep = list(set(range(problem.m)) - set(inds))
        problem.B = problem.B[:, keep_inds]
        problem.d += q0
        problem.k = len(inds)
        print('final inds', len(problem.ind), problem.ind)

    return


def plot_svds(M, tol=None):
    """ Plot a diagram of SVD.

    Parameters
    ----------
    M : :class:`~compas_tno.problems.Problem`
        The problem with matrices for calculation
    tol : float (optional)
        The tolerance to find the independents
    """

    import matplotlib.pyplot as plt
    from numpy.linalg import matrix_rank

    k = len(M.ind)
    n, m = M.E.shape
    # mn_min = min(n, m)

    _, s, _ = svd(asarray(M.E))
    # _, s, _ = svds(M.E, k=min(M.E.shape), solver='propack')
    print('max/min singular vectors E', max(s), min(s), len(s))
    print('Shape E: {} | #ind: {} | rank : {}:'.format(M.E.shape, k, matrix_rank(M.E, tol=tol)))

    # mn = max(m - n, n - m)
    mn = max(m - n, 0)
    zs = k - mn
    print('# null-SVD:', zs)

    # if zs + 3 > len(s):
    #     print('Last {} SVs: {}'.format(len(s[-(zs + 3):]), s[-(zs + 3):]))
    # print('ALL SVS:', s)

    check = check_independents(M)
    print('Check independents:', check)
    print('Shape/ Rank Ed:', M.Ed.shape, matrix_rank(M.Ed))

    if zs > 0:
        first_zero = s[len(s)-zs]
        last_non_zero = s[len(s)-zs - 1]

        x_zs = list(range(len(s)))[-zs:]
        y_zs = s[-zs:]

        print('Last Non Zero:', last_non_zero)
        print('First Zero SV:', first_zero)

        # len_zero = len(s[s<1.0])
        lin_x = len(s) - zs

        porcentage_key = (last_non_zero - first_zero)/last_non_zero
        print('percentage key is:', porcentage_key)

        _, ax = plt.subplots()
        ax.plot(s)
        ax.scatter(x_zs, y_zs)
        ax.plot([lin_x, lin_x], [0, 1.0], color='black')
        ax.plot()
        plt.show()

    else:
        print('EQ. Matrix is FULL rank: (k = m-n)')
        _, ax = plt.subplots()
        # ax.set_ylim(2.7, 3.0)
        ax.plot(s)
        plt.show()

    return



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino


__commandname__ = "TNO_apply_envelope"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']
    proxy = sc.sticky['TNO']['proxy']

    objects = scene.find_by_name('Form')
    if not objects:
        compas_rhino.display_message("There is no FormDiagram in the scene.")
        return
    form = objects[0]

    objects = scene.find_by_name('Shape')
    if not objects:
        compas_rhino.display_message("There is no Shape in the scene.")
        return
    shape = objects[0]

    proxy.package = 'compas_tno.utilities'

    formdata = form.diagram.to_data()
    shapedata = shape.shape.to_data()
    form.diagram.data = proxy.apply_envelope_from_shape_proxy(formdata, shapedata)

    form.settings['show.vertices_bound'] = True
    scene.update()

    validate = compas_rhino.rs.GetString("Validade Envelope", "True", ["True", "False", "Cancel"])
    if validate == "True":
        form.settings['show.vertices_bound'] = False
        scene.update()
    else:
        return

    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



"""
********************************************************************************
compas_tno.algorithms
********************************************************************************

.. currentmodule:: compas_tno.algorithms


Equilibrium
===========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    equilibrium_fdm
    vertical_equilibrium_fdm
    q_from_qid
    q_from_variables
    xyz_from_q
    xyz_from_xopt
    weights_from_xyz
    weights_from_xyz_dict
    compute_reactions
    equilibrium_residual


Independents
============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    find_independents_backward
    find_independents_forward
    find_independents_QR
    find_independents
    independents_exclude
    independents_include
    inds_incl_excl
    check_independents
    check_horizontal_loads


Smoothing
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    constrained_smoothing
    apply_sag


Graphic Statics
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    form_update_with_parallelisation
    force_update_from_form
    reciprocal_from_form


"""
from __future__ import absolute_import

import compas

if not compas.IPY:
    from .equilibrium import (
        equilibrium_fdm,
        vertical_equilibrium_fdm,
        q_from_qid,
        q_from_variables,
        xyz_from_q,
        compute_reactions,
        xyz_from_xopt,
        weights_from_xyz,
        weights_from_xyz_dict,
        equilibrium_residual
    )
    from .independents import (
        find_independents_forward,
        find_independents_backward,
        find_independents_QR,
        find_independents,
        independents_exclude,
        independents_include,
        inds_incl_excl,
        check_independents,
        check_horizontal_loads
    )
    from .smoothing import (
        constrained_smoothing,
        apply_sag
    )
    from .graphstatics import (
        form_update_with_parallelisation,
        force_update_from_form,
        reciprocal_from_form
    )
    # from .equilibrium_pytorch import *  # noqa: F401 F403


__all__ = [
    'equilibrium_fdm',
    'vertical_equilibrium_fdm',
    'q_from_qid',
    'q_from_variables',
    'xyz_from_q',
    'compute_reactions',
    'xyz_from_xopt',
    'weights_from_xyz',
    'weights_from_xyz_dict',
    'equilibrium_residual',

    'find_independents_forward',
    'find_independents_backward',
    'find_independents_QR',
    'find_independents',
    'independents_exclude',
    'independents_include',
    'inds_incl_excl',
    'check_independents',
    'check_horizontal_loads',

    'constrained_smoothing',
    'apply_sag',

    'form_update_with_parallelisation',
    'force_update_from_form',
    'reciprocal_from_form'
]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from scipy.sparse import diags
from scipy.linalg import solve
from scipy.sparse.linalg import spsolve
import numpy as np
import logging
import os
import time


"""
Orginal work Written by Krister Svanberg in Matlab. This is the python version of the code written
by Arjen Deetman.

"""


def mma_numpy(f_g_eval, f_df_g_dg_eval, x0, bounds, args=(), kkttol=1e-6, maxoutit=100, plot='2'):
    """Method of Moving Asymptotes (MMA) implementation from K. Svanberg.

    Parameters
    ----------
    f_g_eval : callable
        The function used by the minimizer to evaluate the objective function (f)
        and the constraints (g). It takes as input the point to evaluate (x), and the
        arguments (args). It returns two arrays, one for the evaluation of (f-1x1 array)
        and one for the evaluation of (g-mx1 array).
    f_df_g_dg_eval : callable
        The function used by the minimizer to evaluate the objective function (f)
        and the constraints (g) and its derivatives. It takes as input the point to evaluate
        (x), and the arguments (args). It returns four arrays, first the evaluation of
        (f - 1x1 array) and (dfdx - nx1 array), and also the evaluation of (g - mx1 array)
        and its derivative (dgdx - mxn array).
    x0 :  array
        Starting point for the MMA optimisation.
    bounds: list
        Bounds for each variable of the optimisation.
    args: tuple
        Argument to be passed with the callables.
    kkttol : float [10e-6]
        The tolerance for the KKT check, i.e. the tolerance for the stopping point observing
        the null derivatives.
    maxoutit : int [100]
        Maximum number of iterations to perform in the optimisation.
    plot : int [2]
        Control plots on screen. '0' nothing is plot, '1' detailled plot and '2' summary plot.

    Returns
    -------
    fopt : float
        The value of the minimisation.
    xval : array
        Value of the variables on the optimum.

    Notes
    -----
    For more info, see [1]. This is the python version of the code written by Arjen Deetman [2].

    Notes
    ----------
    .. [1] Svanberg, K., *The method of moving asymptotes–a new method for structural optimization*,
           International Journal for Numerical Methods in Engineering, 24, 359-373, 1987.
    .. [2] Deetman A. *GCMMA-MMA-Python*. https://github.com/arjendeetman/GCMMA-MMA-Python

    """

    lower = np.array([lw[0] for lw in bounds]).reshape(-1, 1)
    upper = np.array([up[1] for up in bounds]).reshape(-1, 1)

    # Start Time
    start_time = time.time()

    # Logger
    path = os.path.dirname(os.path.realpath(__file__))
    file = os.path.join(path, "GCMMA_TEST.log")
    logger = setup_logger(file)
    logger.info("Started\n")
    # Set numpy print options
    np.set_printoptions(precision=4, formatter={'float': '{: 0.4f}'.format})
    # Beam initial settings
    f0val, g0 = f_g_eval(x0, *args)
    m = len(g0)
    n = len(x0)
    epsimin = 0.0000001
    eeem = np.ones((m, 1))
    zerom = np.zeros((m, 1))
    xval = x0
    xold1 = xval.copy()
    xold2 = xval.copy()
    xmin = lower  # eeen.copy()
    xmax = upper  # 10*eeen
    low = xmin.copy()
    upp = xmax.copy()
    c = 1000*eeem
    d = eeem.copy()
    a0 = 1
    a = zerom.copy()
    raa0 = 0.01
    raa = 0.01*eeem
    raa0eps = 0.000001
    raaeps = 0.000001*eeem
    outeriter = 0
    kkttol = kkttol
    maxoutit = maxoutit

    # Calculate function values and gradients of the objective and constraints functions
    if outeriter == 0:
        f0val, df0dx, fval, dfdx = f_df_g_dg_eval(xval, *args)
        innerit = 0
        outvector1 = np.array([outeriter, innerit, f0val, fval])
        outvector2 = xval.flatten()
        # Log
        if plot == '0':
            pass
        elif plot == '2':
            logger.info("Starting Objective Value = {}\n".format(f0val))
        else:
            logger.info("outvector1 = {}".format(outvector1))
            logger.info("outvector2 = {}\n".format(outvector2))
    # The iterations starts
    kktnorm = kkttol + 10
    outit = 0
    while (kktnorm > kkttol) and (outit < maxoutit):
        outit += 1
        outeriter += 1
        if plot != '0':
            logger.info("iteration = {}".format(outit))
        # The parameters low, upp, raa0 and raa are calculated:
        low, upp, raa0, raa = \
            asymp(outeriter, n, xval, xold1, xold2, xmin, xmax, low, upp, raa0, raa, raa0eps, raaeps, df0dx, dfdx)
        # The MMA subproblem is solved at the point xval:
        xmma, ymma, zmma, lam, xsi, eta, mu, zet, s, f0app, fapp = \
            gcmmasub(m, n, iter, epsimin, xval, xmin, xmax, low, upp,
                     raa0, raa, f0val, df0dx, fval, dfdx, a0, a, c, d)
        # The user should now calculate function values (no gradients) of the objective- and constraint
        # functions at the point xmma ( = the optimal solution of the subproblem).
        f0valnew, fvalnew = f_g_eval(xmma, *args)
        # It is checked if the approximations are conservative:
        conserv = concheck(m, epsimin, f0app, f0valnew, fapp, fvalnew)
        # While the approximations are non-conservative (conserv=0), repeated inner iterations are made:
        innerit = 0
        if conserv == 0:
            while conserv == 0 and innerit <= 15:
                innerit += 1
                # New values on the parameters raa0 and raa are calculated:
                raa0, raa = raaupdate(xmma, xval, xmin, xmax, low, upp, f0valnew, fvalnew, f0app, fapp, raa0,
                                      raa, raa0eps, raaeps, epsimin)
                # The GCMMA subproblem is solved with these new raa0 and raa:
                xmma, ymma, zmma, lam, xsi, eta, mu, zet, s, f0app, fapp = gcmmasub(m, n, iter, epsimin, xval, xmin,
                                                                                    xmax, low, upp, raa0, raa, f0val, df0dx, fval, dfdx, a0, a, c, d)
                # The user should now calculate function values (no gradients) of the objective- and
                # constraint functions at the point xmma ( = the optimal solution of the subproblem).
                f0valnew, fvalnew = f_g_eval(xmma, *args)
                # It is checked if the approximations have become conservative:
                conserv = concheck(m, epsimin, f0app, f0valnew, fapp, fvalnew)
        # Some vectors are updated:
        xold2 = xold1.copy()
        xold1 = xval.copy()
        xval = xmma.copy()
        # Re-calculate function values and gradients of the objective and constraints functions
        f0val, df0dx, fval, dfdx = f_df_g_dg_eval(xval, *args)
        # The residual vector of the KKT conditions is calculated
        residu, kktnorm, residumax = \
            kktcheck(m, n, xmma, ymma, zmma, lam, xsi, eta, mu, zet,
                     s, xmin, xmax, df0dx, fval, dfdx, a0, a, c, d)
        outvector1 = np.array([outeriter, innerit, f0val, fval])
        outvector2 = xval.flatten()
        # Log
        if plot == '0':
            pass
        elif plot == '2':
            logger.info("kktnorm    = {}".format(kktnorm))
            logger.info("Objective Iteration: {}\n".format(f0val))
        else:
            logger.info("outvector1 = {}".format(outvector1))
            logger.info("outvector2 = {}".format(outvector2))
            logger.info("kktnorm    = {}".format(kktnorm))
            logger.info("Objective Iteration: {}\n".format(f0val))
    # Final log
    elapsed_time = time.time() - start_time
    logger.info("Finished")
    if kktnorm < kkttol:
        logger.info('Optimisation Terminated Sucessfully.')
        exitflag = 0
    else:
        logger.info('Optimisation did not respect KKT conditions.')
        exitflag = 1
        if outit == maxoutit:
            logger.info('Maximum iterations reached.')
            exitflag = 2
    logger.info("Objective Value: {}".format(f0val))
    logger.info("Total Iterations: {}".format(outeriter))
    print('Elapsed Time: {0:.1f} sec'.format(elapsed_time))

    return f0val, xval, exitflag


def mmasub(m, n, iter, xval, xmin, xmax, xold1, xold2, f0val, df0dx, fval, dfdx, low, upp, a0, a, c, d, move):
    """
    This function mmasub performs one MMA-iteration, aimed at solving the nonlinear programming problem:

    Minimize    f_0(x) + a_0*z + sum( c_i*y_i + 0.5*d_i*(y_i)^2 )
    subject to  f_i(x) - a_i*z - y_i <= 0,  i = 1,...,m
                xmin_j <= x_j <= xmax_j,    j = 1,...,n
                z >= 0,   y_i >= 0,         i = 1,...,m
    INPUT:

        m     = The number of general constraints.
        n     = The number of variables x_j.
        iter  = Current iteration number ( =1 the first time mmasub is called).
        xval  = Column vector with the current values of the variables x_j.
        xmin  = Column vector with the lower bounds for the variables x_j.
        xmax  = Column vector with the upper bounds for the variables x_j.
        xold1 = xval, one iteration ago (provided that iter>1).
        xold2 = xval, two iterations ago (provided that iter>2).
        f0val = The value of the objective function f_0 at xval.
        df0dx = Column vector with the derivatives of the objective function
                f_0 with respect to the variables x_j, calculated at xval.
        fval  = Column vector with the values of the constraint functions f_i, calculated at xval.
        dfdx  = (m x n)-matrix with the derivatives of the constraint functions
                f_i with respect to the variables x_j, calculated at xval.
                dfdx(i,j) = the derivative of f_i with respect to x_j.
        low   = Column vector with the lower asymptotes from the previous
                iteration (provided that iter>1).
        upp   = Column vector with the upper asymptotes from the previous
                iteration (provided that iter>1).
        a0    = The constants a_0 in the term a_0*z.
        a     = Column vector with the constants a_i in the terms a_i*z.
        c     = Column vector with the constants c_i in the terms c_i*y_i.
        d     = Column vector with the constants d_i in the terms 0.5*d_i*(y_i)^2.

    OUTPUT:

        xmma  = Column vector with the optimal values of the variables x_j
                in the current MMA subproblem.
        ymma  = Column vector with the optimal values of the variables y_i
                in the current MMA subproblem.
        zmma  = Scalar with the optimal value of the variable z
                in the current MMA subproblem.
        lam   = Lagrange multipliers for the m general MMA constraints.
        xsi   = Lagrange multipliers for the n constraints alfa_j - x_j <= 0.
        eta   = Lagrange multipliers for the n constraints x_j - beta_j <= 0.
        mu    = Lagrange multipliers for the m constraints -y_i <= 0.
        zet   = Lagrange multiplier for the single constraint -z <= 0.
        s     = Slack variables for the m general MMA constraints.
        low   = Column vector with the lower asymptotes, calculated and used
                in the current MMA subproblem.
        upp   = Column vector with the upper asymptotes, calculated and used
                in the current MMA subproblem.
    """

    epsimin = 0.0000001
    raa0 = 0.00001
    albefa = 0.1
    asyinit = 0.5
    asyincr = 1.2
    asydecr = 0.7
    eeen = np.ones((n, 1))
    eeem = np.ones((m, 1))
    zeron = np.zeros((n, 1))
    # Calculation of the asymptotes low and upp
    if iter <= 2:
        low = xval-asyinit*(xmax-xmin)
        upp = xval+asyinit*(xmax-xmin)
    else:
        zzz = (xval-xold1)*(xold1-xold2)
        factor = eeen.copy()
        factor[np.where(zzz > 0)] = asyincr
        factor[np.where(zzz < 0)] = asydecr
        low = xval-factor*(xold1-low)
        upp = xval+factor*(upp-xold1)
        lowmin = xval-10*(xmax-xmin)
        lowmax = xval-0.01*(xmax-xmin)
        uppmin = xval+0.01*(xmax-xmin)
        uppmax = xval+10*(xmax-xmin)
        low = np.maximum(low, lowmin)
        low = np.minimum(low, lowmax)
        upp = np.minimum(upp, uppmax)
        upp = np.maximum(upp, uppmin)
    # Calculation of the bounds alfa and beta
    zzz1 = low+albefa*(xval-low)
    zzz2 = xval-move*(xmax-xmin)
    zzz = np.maximum(zzz1, zzz2)
    alfa = np.maximum(zzz, xmin)
    zzz1 = upp-albefa*(upp-xval)
    zzz2 = xval+move*(xmax-xmin)
    zzz = np.minimum(zzz1, zzz2)
    beta = np.minimum(zzz, xmax)
    # Calculations of p0, q0, P, Q and b
    xmami = xmax-xmin
    xmamieps = 0.00001*eeen
    xmami = np.maximum(xmami, xmamieps)
    xmamiinv = eeen/xmami
    ux1 = upp-xval
    ux2 = ux1*ux1
    xl1 = xval-low
    xl2 = xl1*xl1
    uxinv = eeen/ux1
    xlinv = eeen/xl1
    p0 = zeron.copy()
    q0 = zeron.copy()
    p0 = np.maximum(df0dx, 0)
    q0 = np.maximum(-df0dx, 0)
    pq0 = 0.001*(p0+q0)+raa0*xmamiinv
    p0 = p0+pq0
    q0 = q0+pq0
    p0 = p0*ux2
    q0 = q0*xl2
    P = np.zeros((m, n))  # @@ make sparse with scipy?
    Q = np.zeros((m, n))  # @@ make sparse with scipy?
    P = np.maximum(dfdx, 0)
    Q = np.maximum(-dfdx, 0)
    PQ = 0.001*(P+Q)+raa0*np.dot(eeem, xmamiinv.T)
    P = P+PQ
    Q = Q+PQ
    P = (diags(ux2.flatten(), 0).dot(P.T)).T
    Q = (diags(xl2.flatten(), 0).dot(Q.T)).T
    b = (np.dot(P, uxinv)+np.dot(Q, xlinv)-fval)
    # Solving the subproblem by a primal-dual Newton method
    xmma, ymma, zmma, lam, xsi, eta, mu, zet, s = subsolv(m, n, epsimin, low, upp, alfa, beta, p0, q0, P, Q, a0, a, b, c, d)
    # Return values
    return xmma, ymma, zmma, lam, xsi, eta, mu, zet, s, low, upp

# Function for the GCMMA sub problem


def gcmmasub(m, n, iter, epsimin, xval, xmin, xmax, low, upp, raa0, raa, f0val, df0dx, fval, dfdx, a0, a, c, d):
    eeen = np.ones((n, 1))
    zeron = np.zeros((n, 1))
    # Calculations of the bounds alfa and beta
    albefa = 0.1
    zzz = low+albefa*(xval-low)
    alfa = np.maximum(zzz, xmin)
    zzz = upp-albefa*(upp-xval)
    beta = np.minimum(zzz, xmax)
    # Calculations of p0, q0, r0, P, Q, r and b.
    xmami = xmax-xmin
    xmamieps = 0.00001*eeen
    xmami = np.maximum(xmami, xmamieps)
    xmamiinv = eeen/xmami
    ux1 = upp-xval
    ux2 = ux1*ux1
    xl1 = xval-low
    xl2 = xl1*xl1
    uxinv = eeen/ux1
    xlinv = eeen/xl1
    #
    p0 = zeron.copy()
    q0 = zeron.copy()
    p0 = np.maximum(df0dx, 0)
    q0 = np.maximum(-df0dx, 0)
    pq0 = p0+q0
    p0 = p0+0.001*pq0
    q0 = q0+0.001*pq0
    p0 = p0+raa0*xmamiinv
    q0 = q0+raa0*xmamiinv
    p0 = p0*ux2
    q0 = q0*xl2
    r0 = f0val-np.dot(p0.T, uxinv)-np.dot(q0.T, xlinv)
    #
    P = np.zeros((m, n))  # @@ make sparse with scipy?
    Q = np.zeros((m, n))  # @@ make sparse with scipy
    P = (diags(ux2.flatten(), 0).dot(P.T)).T
    Q = (diags(xl2.flatten(), 0).dot(Q.T)).T
    b = (np.dot(P, uxinv)+np.dot(Q, xlinv)-fval)
    P = np.maximum(dfdx, 0)
    Q = np.maximum(-dfdx, 0)
    PQ = P+Q
    P = P+0.001*PQ
    Q = Q+0.001*PQ
    P = P+np.dot(raa, xmamiinv.T)
    Q = Q+np.dot(raa, xmamiinv.T)
    P = (diags(ux2.flatten(), 0).dot(P.T)).T
    Q = (diags(xl2.flatten(), 0).dot(Q.T)).T
    r = fval-np.dot(P, uxinv)-np.dot(Q, xlinv)
    b = -r
    # Solving the subproblem by a primal-dual Newton method
    xmma, ymma, zmma, lam, xsi, eta, mu, zet, s = subsolv(m, n, epsimin, low, upp, alfa, beta, p0, q0, P, Q, a0, a, b, c, d)
    # Calculations of f0app and fapp.
    ux1 = upp-xmma
    xl1 = xmma-low
    uxinv = eeen/ux1
    xlinv = eeen/xl1
    f0app = r0+np.dot(p0.T, uxinv)+np.dot(q0.T, xlinv)
    fapp = r+np.dot(P, uxinv)+np.dot(Q, xlinv)
    # Return values
    return xmma, ymma, zmma, lam, xsi, eta, mu, zet, s, f0app, fapp

# Function for solving the subproblem (can be used for MMA and GCMMA)


def subsolv(m, n, epsimin, low, upp, alfa, beta, p0, q0, P, Q, a0, a, b, c, d):
    """
    This function subsolv solves the MMA subproblem:

    minimize SUM[p0j/(uppj-xj) + q0j/(xj-lowj)] + a0*z + SUM[ci*yi + 0.5*di*(yi)^2],

    subject to SUM[pij/(uppj-xj) + qij/(xj-lowj)] - ai*z - yi <= bi,
        alfaj <=  xj <=  betaj,  yi >= 0,  z >= 0.

    Input:  m, n, low, upp, alfa, beta, p0, q0, P, Q, a0, a, b, c, d.
    Output: xmma,ymma,zmma, slack variables and Lagrange multiplers.
    """

    een = np.ones((n, 1))
    eem = np.ones((m, 1))
    epsi = 1
    epsvecn = epsi*een
    epsvecm = epsi*eem
    x = 0.5*(alfa+beta)
    y = eem.copy()
    z = np.array([[1.0]])
    lam = eem.copy()
    xsi = een/(x-alfa)
    xsi = np.maximum(xsi, een)
    eta = een/(beta-x)
    eta = np.maximum(eta, een)
    mu = np.maximum(eem, 0.5*c)
    zet = np.array([[1.0]])
    s = eem.copy()
    itera = 0
    # Start while epsi>epsimin
    while epsi > epsimin:
        epsvecn = epsi*een
        epsvecm = epsi*eem
        ux1 = upp-x
        xl1 = x-low
        ux2 = ux1*ux1
        xl2 = xl1*xl1
        uxinv1 = een/ux1
        xlinv1 = een/xl1
        plam = p0+np.dot(P.T, lam)
        qlam = q0+np.dot(Q.T, lam)
        gvec = np.dot(P, uxinv1)+np.dot(Q, xlinv1)
        dpsidx = plam/ux2-qlam/xl2
        rex = dpsidx-xsi+eta
        rey = c+d*y-mu-lam
        rez = a0-zet-np.dot(a.T, lam)
        relam = gvec-a*z-y+s-b
        rexsi = xsi*(x-alfa)-epsvecn
        reeta = eta*(beta-x)-epsvecn
        remu = mu*y-epsvecm
        rezet = zet*z-epsi
        res = lam*s-epsvecm
        residu1 = np.concatenate((rex, rey, rez), axis=0)
        residu2 = np.concatenate((relam, rexsi, reeta, remu, rezet, res), axis=0)
        residu = np.concatenate((residu1, residu2), axis=0)
        residunorm = np.sqrt((np.dot(residu.T, residu)).item())
        residumax = np.max(np.abs(residu))
        ittt = 0
        # Start while (residumax>0.9*epsi) and (ittt<200)
        while (residumax > 0.9*epsi) and (ittt < 200):
            ittt = ittt+1
            itera = itera+1
            ux1 = upp-x
            xl1 = x-low
            ux2 = ux1*ux1
            xl2 = xl1*xl1
            ux3 = ux1*ux2
            xl3 = xl1*xl2
            uxinv1 = een/ux1
            xlinv1 = een/xl1
            uxinv2 = een/ux2
            xlinv2 = een/xl2
            plam = p0+np.dot(P.T, lam)
            qlam = q0+np.dot(Q.T, lam)
            gvec = np.dot(P, uxinv1)+np.dot(Q, xlinv1)
            GG = (diags(uxinv2.flatten(), 0).dot(P.T)).T-(diags(xlinv2.flatten(), 0).dot(Q.T)).T
            dpsidx = plam/ux2-qlam/xl2
            delx = dpsidx-epsvecn/(x-alfa)+epsvecn/(beta-x)
            dely = c+d*y-lam-epsvecm/y
            delz = a0-np.dot(a.T, lam)-epsi/z
            dellam = gvec-a*z-y-b+epsvecm/lam
            diagx = plam/ux3+qlam/xl3
            diagx = 2*diagx+xsi/(x-alfa)+eta/(beta-x)
            diagxinv = een/diagx
            diagy = d+mu/y
            diagyinv = eem/diagy
            diaglam = s/lam
            diaglamyi = diaglam+diagyinv
            # Start if m<n
            if m < n:
                blam = dellam+dely/diagy-np.dot(GG, (delx/diagx))
                bb = np.concatenate((blam, delz), axis=0)
                Alam = np.asarray(diags(diaglamyi.flatten(), 0)
                                  + (diags(diagxinv.flatten(), 0).dot(GG.T).T).dot(GG.T))
                AAr1 = np.concatenate((Alam, a), axis=1)
                AAr2 = np.concatenate((a, -zet/z), axis=0).T
                AA = np.concatenate((AAr1, AAr2), axis=0)
                solut = spsolve(AA, bb)
                # solut = solve(AA, bb)
                dlam = solut[0:m]
                dz = solut[m:m+1]
                dx = -delx/diagx-np.dot(GG.T, dlam)/diagx
            else:
                diaglamyiinv = eem/diaglamyi
                dellamyi = dellam+dely/diagy
                Axx = np.asarray(diags(diagx.flatten(), 0)
                                 + (diags(diaglamyiinv.flatten(), 0).dot(GG).T).dot(GG))
                azz = zet/z+np.dot(a.T, (a/diaglamyi))
                axz = np.dot(-GG.T, (a/diaglamyi))
                bx = delx+np.dot(GG.T, (dellamyi/diaglamyi))
                bz = delz-np.dot(a.T, (dellamyi/diaglamyi))
                AAr1 = np.concatenate((Axx, axz), axis=1)
                AAr2 = np.concatenate((axz.T, azz), axis=1)
                AA = np.concatenate((AAr1, AAr2), axis=0)
                bb = np.concatenate((-bx, -bz), axis=0)
                # print(AA.shape)
                # print(bb.shape)
                # print(AA)
                # print(bb)
                solut = solve(AA, bb)
                # solut = spsolve(AA, bb)
                dx = solut[0:n]
                dz = solut[n:n+1]
                dlam = np.dot(GG, dx)/diaglamyi-dz*(a/diaglamyi)+dellamyi/diaglamyi
                # End if m<n
            dy = -dely/diagy+dlam/diagy
            dxsi = -xsi+epsvecn/(x-alfa)-(xsi*dx)/(x-alfa)
            deta = -eta+epsvecn/(beta-x)+(eta*dx)/(beta-x)
            dmu = -mu+epsvecm/y-(mu*dy)/y
            dzet = -zet+epsi/z-zet*dz/z
            ds = -s+epsvecm/lam-(s*dlam)/lam
            xx = np.concatenate((y, z, lam, xsi, eta, mu, zet, s), axis=0)
            dxx = np.concatenate((dy, dz, dlam, dxsi, deta, dmu, dzet, ds), axis=0)
            #
            stepxx = -1.01*dxx/xx
            stmxx = np.max(stepxx)
            stepalfa = -1.01*dx/(x-alfa)
            stmalfa = np.max(stepalfa)
            stepbeta = 1.01*dx/(beta-x)
            stmbeta = np.max(stepbeta)
            stmalbe = max(stmalfa, stmbeta)
            stmalbexx = max(stmalbe, stmxx)
            stminv = max(stmalbexx, 1.0)
            steg = 1.0/stminv
            #
            xold = x.copy()
            yold = y.copy()
            zold = z.copy()
            lamold = lam.copy()
            xsiold = xsi.copy()
            etaold = eta.copy()
            muold = mu.copy()
            zetold = zet.copy()
            sold = s.copy()
            #
            itto = 0
            resinew = 2*residunorm
            # Start: while (resinew>residunorm) and (itto<50)
            while (resinew > residunorm) and (itto < 50):
                itto = itto+1
                x = xold+steg*dx
                y = yold+steg*dy
                z = zold+steg*dz
                lam = lamold+steg*dlam
                xsi = xsiold+steg*dxsi
                eta = etaold+steg*deta
                mu = muold+steg*dmu
                zet = zetold+steg*dzet
                s = sold+steg*ds
                ux1 = upp-x
                xl1 = x-low
                ux2 = ux1*ux1
                xl2 = xl1*xl1
                uxinv1 = een/ux1
                xlinv1 = een/xl1
                plam = p0+np.dot(P.T, lam)
                qlam = q0+np.dot(Q.T, lam)
                gvec = np.dot(P, uxinv1)+np.dot(Q, xlinv1)
                dpsidx = plam/ux2-qlam/xl2
                rex = dpsidx-xsi+eta
                rey = c+d*y-mu-lam
                rez = a0-zet-np.dot(a.T, lam)
                relam = gvec-np.dot(a, z)-y+s-b
                rexsi = xsi*(x-alfa)-epsvecn
                reeta = eta*(beta-x)-epsvecn
                remu = mu*y-epsvecm
                rezet = np.dot(zet, z)-epsi
                res = lam*s-epsvecm
                residu1 = np.concatenate((rex, rey, rez), axis=0)
                residu2 = np.concatenate((relam, rexsi, reeta, remu, rezet, res), axis=0)
                residu = np.concatenate((residu1, residu2), axis=0)
                resinew = np.sqrt(np.dot(residu.T, residu))
                steg = steg/2
                # End: while (resinew>residunorm) and (itto<50)
            residunorm = resinew.copy()
            residumax = max(abs(residu))
            steg = 2*steg
            # End: while (residumax>0.9*epsi) and (ittt<200)
        epsi = 0.1*epsi
        # End: while epsi>epsimin
    xmma = x.copy()
    ymma = y.copy()
    zmma = z.copy()
    lamma = lam
    xsimma = xsi
    etamma = eta
    mumma = mu
    zetmma = zet
    smma = s
    # Return values
    return xmma, ymma, zmma, lamma, xsimma, etamma, mumma, zetmma, smma

# Function for Karush–Kuhn–Tucker check


def kktcheck(m, n, x, y, z, lam, xsi, eta, mu, zet, s, xmin, xmax, df0dx, fval, dfdx, a0, a, c, d):
    """
    The left hand sides of the KKT conditions for the following nonlinear programming problem are
    calculated.

    Minimize f_0(x) + a_0*z + sum(c_i*y_i + 0.5*d_i*(y_i)^2)
    subject to  f_i(x) - a_i*z - y_i <= 0,   i = 1,...,m
                xmax_j <= x_j <= xmin_j,     j = 1,...,n
                z >= 0,   y_i >= 0,          i = 1,...,m

    INPUT:

        m     = The number of general constraints.
        n     = The number of variables x_j.
        x     = Current values of the n variables x_j.
        y     = Current values of the m variables y_i.
        z     = Current value of the single variable z.
        lam   = Lagrange multipliers for the m general constraints.
        xsi   = Lagrange multipliers for the n constraints xmin_j - x_j <= 0.
        eta   = Lagrange multipliers for the n constraints x_j - xmax_j <= 0.
        mu    = Lagrange multipliers for the m constraints -y_i <= 0.
        zet   = Lagrange multiplier for the single constraint -z <= 0.
        s     = Slack variables for the m general constraints.
        xmin  = Lower bounds for the variables x_j.
        xmax  = Upper bounds for the variables x_j.
        df0dx = Vector with the derivatives of the objective function f_0
                with respect to the variables x_j, calculated at x.
        fval  = Vector with the values of the constraint functions f_i,
                calculated at x.
        dfdx  = (m x n)-matrix with the derivatives of the constraint functions
                f_i with respect to the variables x_j, calculated at x.
                dfdx(i,j) = the derivative of f_i with respect to x_j.
        a0    = The constants a_0 in the term a_0*z.
        a     = Vector with the constants a_i in the terms a_i*z.
        c     = Vector with the constants c_i in the terms c_i*y_i.
        d     = Vector with the constants d_i in the terms 0.5*d_i*(y_i)^2.

    OUTPUT:

        residu     = the residual vector for the KKT conditions.
        residunorm = sqrt(residu'*residu).
        residumax  = max(abs(residu)).

    """

    rex = df0dx+np.dot(dfdx.T, lam)-xsi+eta
    rey = c+d*y-mu-lam
    rez = a0-zet-np.dot(a.T, lam)
    relam = fval-a*z-y+s
    rexsi = xsi*(x-xmin)
    reeta = eta*(xmax-x)
    remu = mu*y
    rezet = zet*z
    res = lam*s
    residu1 = np.concatenate((rex, rey, rez), axis=0)
    residu2 = np.concatenate((relam, rexsi, reeta, remu, rezet, res), axis=0)
    residu = np.concatenate((residu1, residu2), axis=0)
    residunorm = np.sqrt((np.dot(residu.T, residu)).item())
    residumax = np.max(np.abs(residu))
    return residu, residunorm, residumax

# Function for updating raa0 and raa


def raaupdate(xmma, xval, xmin, xmax, low, upp, f0valnew, fvalnew, f0app, fapp, raa0, raa, raa0eps, raaeps, epsimin):
    """
    Values of the parameters raa0 and raa are updated during an inner iteration.
    """

    raacofmin = 1e-12
    eeem = np.ones((raa.size, 1))
    eeen = np.ones((xmma.size, 1))
    xmami = xmax-xmin
    xmamieps = 0.00001*eeen
    xmami = np.maximum(xmami, xmamieps)
    xxux = (xmma-xval)/(upp-xmma)
    xxxl = (xmma-xval)/(xmma-low)
    xxul = xxux*xxxl
    ulxx = (upp-low)/xmami
    raacof = np.dot(xxul.T, ulxx)
    raacof = np.maximum(raacof, raacofmin)
    #
    f0appe = f0app+0.5*epsimin
    if np.all(f0valnew > f0appe):
        deltaraa0 = (1.0/raacof)*(f0valnew-f0app)
        zz0 = 1.1*(raa0+deltaraa0)
        zz0 = np.minimum(zz0, 10*raa0)
        raa0 = zz0
    #
    fappe = fapp+0.5*epsimin*eeem
    fdelta = fvalnew-fappe
    deltaraa = (1/raacof)*(fvalnew-fapp)
    zzz = 1.1*(raa+deltaraa)
    zzz = np.minimum(zzz, 10*raa)
    raa[np.where(fdelta > 0)] = zzz[np.where(fdelta > 0)]
    #
    return raa0, raa

# Function to check if the approsimations are conservative


def concheck(m, epsimin, f0app, f0valnew, fapp, fvalnew):
    """
    If the current approximations are conservative, the parameter conserv is set to 1.
    """

    eeem = np.ones((m, 1))
    f0appe = f0app+epsimin
    fappe = fapp+epsimin*eeem
    arr1 = np.concatenate((f0appe.flatten(), fappe.flatten()))
    arr2 = np.concatenate((f0valnew.flatten(), fvalnew.flatten()))
    if np.all(arr1 >= arr2):
        conserv = 1
    else:
        conserv = 0
    return conserv

# Calculate low, upp, raa0, raa in the beginning of each outer iteration


def asymp(outeriter, n, xval, xold1, xold2, xmin, xmax, low, upp, raa0, raa, raa0eps, raaeps, df0dx, dfdx):
    """
    Values on the parameters raa0, raa, low and upp are calculated in the beginning of each outer
    iteration.
    """

    eeen = np.ones((n, 1))
    asyinit = 0.5
    asyincr = 1.2
    asydecr = 0.7
    xmami = xmax-xmin
    xmamieps = 0.00001*eeen
    xmami = np.maximum(xmami, xmamieps)
    raa0 = np.dot(np.abs(df0dx).T, xmami)
    raa0 = np.maximum(raa0eps, (0.1/n)*raa0)
    raa = np.dot(np.abs(dfdx), xmami)
    raa = np.maximum(raaeps, (0.1/n)*raa)
    if outeriter <= 2:
        low = xval-asyinit*xmami
        upp = xval+asyinit*xmami
    else:
        xxx = (xval-xold1)*(xold1-xold2)
        factor = eeen.copy()
        factor[np.where(xxx > 0)] = asyincr
        factor[np.where(xxx < 0)] = asydecr
        low = xval-factor*(xold1-low)
        upp = xval+factor*(upp-xold1)
        lowmin = xval-10*xmami
        lowmax = xval-0.01*xmami
        uppmin = xval+0.01*xmami
        uppmax = xval+10*xmami
        low = np.maximum(low, lowmin)
        low = np.minimum(low, lowmax)
        upp = np.minimum(upp, uppmax)
        upp = np.maximum(upp, uppmin)
    return low, upp, raa0, raa


def setup_logger(logfile):
    # Create logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    # Create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    # Create file handler and set level to debug
    fh = logging.FileHandler(logfile)
    fh.setLevel(logging.DEBUG)
    # Add formatter to ch and fh
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    fh.setFormatter(formatter)
    # Add ch and fh to logger
    logger.addHandler(ch)
    logger.addHandler(fh)
    # Open logfile and reset
    with open(logfile, 'w'):
        pass
    # Return logger
    return logger


# Function for the example with evaluation of f and g
def _evaluate_f_g(xval, *args):
    nx = 5
    eeen = np.ones((nx, 1))
    c1 = args[5]
    c2 = args[6]
    aaa = np.array([args[:5]]).T
    xval2 = xval*xval
    xval3 = xval2*xval
    xinv3 = eeen/xval3
    f0val = c1*np.dot(eeen.T, xval)
    fval = np.dot(aaa.T, xinv3)-c2
    return f0val, fval

# Function for the example with evaluation of f and g


def _evaluate_f_g_derivatives(xval, *args):
    nx = 5
    eeen = np.ones((nx, 1))
    c1 = args[5]
    c2 = args[6]
    aaa = np.array([args[:5]]).T
    xval2 = xval*xval
    xval3 = xval2*xval
    xval4 = xval2*xval2
    xinv3 = eeen/xval3
    xinv4 = eeen/xval4
    f0val = c1*np.dot(eeen.T, xval).item()
    df0dx = c1*eeen
    fval = np.dot(aaa.T, xinv3).item()-c2
    dfdx = -3*(aaa*xinv4).T
    return f0val, df0dx, fval, dfdx

# ==============================================================================
# Main
# ==============================================================================


if __name__ == "__main__":

    """
    Example of the "beam problem" from the MMA paper of Krister Svanberg.

        minimize 0.0624*(x(1) + x(2) + x(3) + x(4) + x(5))
        subject to 61/(x(1)^3) + 37/(x(2)^3) + 19/(x(3)^3) + 7/(x(4)^3) + 1/(x(5)^3) =< 1,
                1 =< x(j) =< 10, for j=1,..,5.
    """

    # Set starting point, bounds and additional arguments for the objective and constrainnt functions
    x0 = 5 * np.ones((5, 1))
    bounds = [[1, 10]*5]
    args = (61.0, 37.0, 19.0, 7.0, 1.0, 0.0624, 1)

    # Call numpy MMA and retrieve the optimum value, optimum solution and exitflag.
    fopt, xopt, exitflag = mma_numpy(_evaluate_f_g, _evaluate_f_g_derivatives, x0, bounds, args)

    print('\nOptimum Solution for Beam problem: ')
    print(xopt)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import scriptcontext as sc

import compas_rhino
from compas_rhino.geometry import RhinoMesh
from compas_tno.shapes import Shape


__commandname__ = "TNO_shape_from_meshes"


def RunCommand(is_interactive):

    if 'TNO' not in sc.sticky:
        compas_rhino.display_message('TNO has not been initialised yet.')
        return

    scene = sc.sticky['TNO']['scene']

    answer = compas_rhino.rs.GetString("Select a method to input the shape. From", "Cancel", ["IntraExtrados", "IntraExtradosMiddle", "Middle", "Cancel"])
    if not answer:
        return
    if answer == "Cancel":
        return

    middle = None

    warning = 'Warning: Select only one mesh.'

    if answer == "IntraExtrados" or answer == "IntraExtradosMiddle":
        selection = compas_rhino.select_meshes(message='Select Intrados Mesh')
        if not selection:
            return
        if len(selection) > 1:
            compas_rhino.display_message(warning)
        guid_intrados = selection[0]
        intrados_RM = RhinoMesh.from_guid(guid_intrados)
        intrados = intrados_RM.to_compas()

        compas_rhino.rs.UnselectObjects(selection)

        selection = compas_rhino.select_mesh(message='Select Extrados Mesh')
        if not selection:
            return
        if len(selection) > 1:
            compas_rhino.display_message(warning)
        guid_extrados = selection[0]
        extrados_RM = RhinoMesh.from_guid(guid_extrados)
        extrados = extrados_RM.to_compas()

        compas_rhino.rs.UnselectObjects(selection)

    if answer == "Middle" or answer == "IntraExtradosMiddle":
        selection = compas_rhino.select_mesh(message='Select Middle Mesh')
        if not selection:
            return
        if len(selection) > 1:
            compas_rhino.display_message(warning)
        guid_middle = selection[0]
        middle_RM = RhinoMesh.from_guid(guid_middle)
        middle = middle_RM.to_compas()

    thk = compas_rhino.rs.GetReal("Input thickness value (estimate)", 0.50)
    if not thk:
        return

    shape = Shape.from_meshes(intrados, extrados, middle, data={'type': 'general', 't': 0.0, 'thk': thk})

    scene.add(shape, name='Shape', layer='TNO::Shape')
    scene.update()
    scene.save()


# ==============================================================================
# Main
# ==============================================================================
if __name__ == '__main__':

    RunCommand(True)



"""
********************************************************************************
compas_tno.rhino
********************************************************************************

.. currentmodule:: compas_tno.rhino


"""


from __future__ import absolute_import

import compas

from compas_tno.diagrams import FormDiagram
from compas_tno.diagrams import ForceDiagram
from compas_tno.shapes import Shape
from compas_tno.optimisers import Optimiser

from .diagramartist import DiagramArtist  # noqa: F401
from .formartist import FormArtist  # noqa: F401
from .forceartist import ForceArtist  # noqa: F401
from .shapeartist import ShapeArtist

from .diagramobject import DiagramObject  # noqa: F401
from .formobject import FormObject  # noqa: F401
from .forceobject import ForceObject  # noqa: F401
from .shapeobject import ShapeObject
from .optimiserobject import OptimiserObject  # noqa: F401

from .scene import Scene  # noqa: F401
from .settings import SettingsForm  # noqa: F401
from .forms import AttributesForm  # noqa: F401
from .forms import Browser  # noqa: F401

if compas.IPY:

    from compas_ui.objects import Object
    from compas_rhino.artists import RhinoArtist

    DiagramObject.register(FormDiagram, FormObject)
    RhinoArtist.register(FormDiagram, FormArtist, context='Rhino')

    DiagramObject.register(ForceDiagram, ForceObject)
    RhinoArtist.register(ForceDiagram, ForceArtist, context='Rhino')

    Object.register(Shape, ShapeObject)
    RhinoArtist.register(Shape, ShapeArtist, context='Rhino')

    Object.register(Optimiser, OptimiserObject)
    RhinoArtist.register(Optimiser, DiagramArtist, context='Rhino')

    print('did my registering')

__all__ = [name for name in dir() if not name.startswith('_')]



from compas.datastructures import mesh_dual
from compas.datastructures import Mesh
from compas.geometry import add_vectors, scale_vector


def extended_dual(form, cls=None):
    """Create the extended dual of the mesh, which is the centroid dual added with the faces in the boundary.

    Parameters
    ----------
    form : FormDiagram
        Form Diagram as the primal
    cls : Class, optional
        The class which should return the dual, by default None

    Returns
    -------
    dual
        Extended dual mesh in the class required
    """

    dual = mesh_dual(form, cls)
    dual.flip_cycles()

    edge_vertex = {}

    boundary_edges = [edge for boundary in form.edges_on_boundaries() for edge in boundary]

    for u, v in boundary_edges:
        x, y, z = form.edge_midpoint(u, v)
        edge_vertex[u, v] = edge_vertex[v, u] = dual.add_vertex(x=x, y=y, z=z)

    boundary_vertices = [vertex for boundary in form.vertices_on_boundaries() for vertex in boundary]
    dual_vertices = list(dual.vertices())

    for vertex in boundary_vertices:
        x, y, z = form.vertex_coordinates(vertex)
        vertices = [dual.add_vertex(x=x, y=y, z=z)]
        nbrs = form.vertex_neighbors(vertex, ordered=True)[::-1]

        for nbr in nbrs:
            if form.is_edge_on_boundary(vertex, nbr):
                vertices.append(edge_vertex[vertex, nbr])

            face = form.halfedge[vertex][nbr]
            if face is not None:
                if face in dual_vertices:
                    vertices.append(face)

        dual.add_face(vertices)

    return dual


def blocks_from_dual(dual, thk):
    idos = dual.copy()
    edos = dual.copy()

    from compas_assembly.datastructures import Assembly

    assembly = Assembly()

    for vertex in dual.vertices():
        point = dual.vertex_coordinates(vertex)
        normal = dual.vertex_normal(vertex)

        idos.vertex_attributes(vertex, 'xyz', add_vectors(point, scale_vector(normal, -0.5 * thk)))
        edos.vertex_attributes(vertex, 'xyz', add_vectors(point, scale_vector(normal, 0.5 * thk)))

    faceslist = list(idos.faces())
    for face in idos.faces():
        bottom = idos.face_coordinates(face)
        top = edos.face_coordinates(face)

        f = len(bottom)

        faces = [
            list(range(f)),
            list(range(f + f - 1, f - 1, -1))]

        for i in range(f - 1):
            faces.append([i, i + f, i + f + 1, i + 1])
        faces.append([f - 1, f + f - 1, f, 0])

        block = Mesh.from_vertices_and_faces(bottom + top, faces)
        if face < faceslist[-1]:
            assembly.add_block(block)  # last block is repeated

    return assembly



from numpy import zeros


__all__ = [
    'compute_form_initial_lengths',
    'compute_edge_stiffness',
    'compute_average_edge_stiffness'
]


def compute_form_initial_lengths(form):
    """Compute the length of each edge based on the projection of the pattern onto the middle surface.
    Note: the middle surface is obtained based on the "target" attributes of the form diagram that should
    be set before invoking this method.

    Parameters
    ----------
    form : ::FormDiagram::
        Form diagram to compute the lengths.

    Returns
    -------
    lengths : array (m x 1)
        A vector (m x 1) containing the calculated edge lengths.
    """

    form_ = form.copy()

    m = form.number_of_real_edges()
    lengths = zeros((m, 1))

    for key in form_.vertices():
        z = form.vertex_attribute(key, 'target')
        form_.vertex_attribute(key, 'z', value=z)

    i = 0
    for edge in form.edges():
        lengths[i] = form_.edge_length(*edge)
        form.edge_attribute(edge, 'l0', lengths[i])
        i += 1

    return lengths


def compute_edge_stiffness(form, lengths=None, area=0.125, E=20E+6):
    r"""Compute the stiffness of each edge based on the initial lengths, constant area and Young Modulus E.

    Parameters
    ----------
    form : ::FormDiagram::
        Form diagram to compute the stiffness.
    lengths : array (m x 1)
        The lengths considered in each edge.
        The default value is ``None`` in which case the initial lengths are calculated based on the middle surface.
    area : float
        The constant area applied to each edge.
        The default value is ``0.125`` which corresponds to a square section of 0.25 x 0.25 m.
    E : float
        The young modulus of the material.
        The default value is ``20E+6`` which corresponds to a 20 GPa commonly used for concrete.

    Returns
    -------
    k : array (m x 1)
        The stiffness of each edge.

    Note
    -------
    The stiffness is based on a the formula below:

    $k = \frac{EA}{l_\mathrm{i}}$

    """

    if lengths is None:
        lengths = compute_form_initial_lengths(form)
    m = len(lengths)
    k = zeros((m, 1))

    for index, edge in enumerate(list(form.edges_where({'_is_edge': True}))):
        k[index] = E * area / lengths[index]
        form.edge_attribute(edge, 'k', k[index])

    return k


def compute_average_edge_stiffness(form, E=20E+6, Ah=10E-4):
    """Compute the stiffness divided by length of each edge (constant thoughout the form diagram)
    based on the Young Modulus E and the Axial stress Ah.

    Parameters
    ----------
    form : ::FormDiagram::
        Form diagram to compute the stiffness.
    E : float
        The young modulus of the material.
        The default value is ``20E+6`` which corresponds to a 20 GPa commonly used for concrete.
    Ah : float
        The axial stress which is assumed to be constant throughout the edges.
        The default value is ``10E-4`` which corresponts to a compressive stress of 10 MPa in the section.

    Returns
    -------
    k : float
        The stiffness divided by length on the edges (constant).

    """

    k = E * Ah

    return k



from numpy import sum as npsum
from numpy.linalg import norm

from compas_tno.algorithms import xyz_from_q
from compas_tno.algorithms import q_from_variables

from scipy.sparse import diags
from scipy.sparse.linalg import splu

from compas_tno.algorithms import weights_from_xyz

from compas_tno.problems import gradient_fmin
from compas_tno.problems import gradient_fmax
from compas_tno.problems import gradient_bestfit
from compas_tno.problems import gradient_horprojection
from compas_tno.problems import gradient_loadpath
from compas_tno.problems import gradient_complementary_energy
from compas_tno.problems import gradient_complementary_energy_nonlinear
from compas_tno.problems import gradient_max_section
from compas_tno.problems import gradient_feasibility
from compas_tno.problems import gradient_reduce_thk
from compas_tno.problems import gradient_tight_crosssection


def objective_selector(objective):
    """Select objective callable and gradient vector based on the desired objective function.
    For the complete list of objectives that can be selected see [link].

    Parameters
    ----------
    objective : str
        The name of the objective. See ``Optimiser`` for a complete list of objectivees.

    Returns
    -------
    fobj : callable
        Callable to compute the value of the objective function in the point.
    fgrad : callable
        Callable to compute the gradient of the objective function in the point.

    """

    if objective == 'loadpath':
        fobj = f_loadpath_general
        fgrad = gradient_loadpath
    elif objective == 'target' or objective == 'bestfit':
        fobj = f_bestfit
        fgrad = gradient_bestfit
    elif objective == 'min':
        fobj = f_min_thrust
        fgrad = gradient_fmin
    elif objective == 'max':
        fobj = f_max_thrust
        fgrad = gradient_fmax
    elif objective == 'feasibility':
        fobj = f_constant
        fgrad = gradient_feasibility
    elif objective == 'hor_projection':
        fobj = f_horprojection
        fgrad = gradient_horprojection
    elif objective == 't':  # analytical reduce thickness
        fobj = f_reduce_thk
        fgrad = gradient_reduce_thk
    elif objective == 's':  # tight UB and LB 0 -> 1/2
        fobj = f_tight_crosssection
        fgrad = gradient_tight_crosssection
    elif objective == 'n':  # vector n offset the surfaces -> larger the better (higher GSF)
        fobj = f_tight_crosssection
        fgrad = gradient_tight_crosssection
    elif objective == 'Ecomp-linear':  # vector lambda as hor multiplier larger the better (higher GSF)
        fobj = f_complementary_energy
        fgrad = gradient_complementary_energy
    elif objective == 'Ecomp-nonlinear':
        fobj = f_complementary_energy_nonlinear
        fgrad = gradient_complementary_energy_nonlinear
    elif objective == 'max_section':
        fobj = f_max_section
        fgrad = gradient_max_section
    elif objective == 'max_load':
        fobj = f_tight_crosssection
        fgrad = gradient_tight_crosssection
    else:
        print('Please, provide a valid objective for the optimisation')
        raise NotImplementedError

    return fobj, fgrad


def f_min_thrust(variables, M):
    """Objective function to minimise the horizontal thrust

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    if isinstance(M, list):
        M = M[0]

    update_geometry = False  # In the typical case the geometry does not need to be updated
    k = M.k
    nb = len(M.fixed)

    P_Xh_fixed = M.P[M.fixed][:, :2]  # Horizontal loads in the fixed vertices
    P_free = M.P[M.free]  # Loads in the free vertices

    X = M.X
    Xh = X[:, :2]  # xy of the thrust (i.e. horizontal projection)

    SPLU_D = None

    qid = variables[:k].reshape(-1, 1)
    q = q_from_variables(qid, M.B, M.d)
    Q = diags(q.flatten())

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
        update_geometry = True
    if 'zb' in M.variables:
        zb = variables[-nb:]
        X[M.fixed, 2] = zb.flatten()
        if 'fixed' not in M.features:
            update_geometry = True

    if update_geometry:
        if 'update-loads' in M.features:
            CitQCi = M.Cit @ Q @ M.Ci
            SPLU_D = splu(CitQCi)
            X[M.free] = xyz_from_q(q, P_free, X[M.fixed], M.Ci, M.Cit, M.Cb, SPLU_D=SPLU_D)
            pz = -1 * weights_from_xyz(X, M.F, M.V0, M.V1, M.V2, thk=M.thk, density=M.ro)
            P_free[:, 2] = pz[M.free]

        X[M.free] = xyz_from_q(q, P_free, X[M.fixed], M.Ci, M.Cit, M.Cb, SPLU_D=SPLU_D)
        Xh = X[:, :2]

    CfQC = M.Cb.transpose() @ Q @ M.C
    Rh = CfQC @ Xh - P_Xh_fixed
    f = sum(norm(Rh, axis=1))

    return f


def f_max_thrust(variables, M):
    """Objective function to maximise the horizontal thrust

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    return -1 * f_min_thrust(variables, M)


def f_bestfit(variables, M):
    """Objective function to minimise the vertical squared distance to a given target

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    if isinstance(M, list):
        M = M[0]

    k = M.k
    nb = len(M.fixed)

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    f = sum((M.X[:, [2]] - M.s)**2)

    return f


def f_horprojection(variables, M):
    """Objective function to minimise the horizontal squared distance of the nodes on the form diagram to a given pattern

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    if isinstance(M, list):
        M = M[0]

    k = M.k
    nb = len(M.fixed)

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    f = sum((M.X[:, [0]] - M.x0)**2) + sum((M.X[:, [1]] - M.y0)**2)

    return f


def f_loadpath_general(variables, M):
    """Objective function to minimise the loadpath

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    if isinstance(M, list):
        M = M[0]

    k = M.k
    nb = len(M.fixed)

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    uvw = M.C @ M.X

    l2 = npsum(uvw**2, axis=1).reshape(-1, 1)

    f = (abs(M.q.transpose()).dot(l2)).item()

    return f


def f_complementary_energy(variables, M):
    """Objective function to minimise the complementary energy to a given applied foundation displacement

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    if isinstance(M, list):
        M = M[0]

    k = M.k
    nb = len(M.fixed)

    qid = variables[:k].reshape(-1, 1)
    M.q = q_from_variables(qid, M.B, M.d)

    if 'xyb' in M.variables:
        xyb = variables[k:k + 2*nb]
        M.X[M.fixed, :2] = xyb.reshape(-1, 2, order='F')
    if 'zb' in M.variables:
        zb = variables[-nb:]
        M.X[M.fixed, [2]] = zb.flatten()

    M.X[M.free] = xyz_from_q(M.q, M.P[M.free], M.X[M.fixed], M.Ci, M.Cit, M.Cb)

    CfQC = M.Cb.transpose().dot(diags(M.q.flatten())).dot(M.C)
    R = CfQC.dot(M.X) - M.P[M.fixed]
    f = -1 * npsum(R*M.dXb)

    return f


def f_complementary_energy_nonlinear(variables, M):
    """Objective function to minimise the nonlinear complementary energy to a given applied foundation displacement

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    if isinstance(M, list):
        M = M[0]

    flin = f_complementary_energy(variables, M)

    if M.Ecomp_method == 'simplified':
        fquad = npsum(M.stiff * M.q.reshape(-1, 1) ** 2)  # assuming area and lengths constant - computed in beginning
    if M.Ecomp_method == 'complete':
        fquad = f_loadpath_general(variables, M) * M.stiff

    f = flin + fquad

    return f


def f_max_section(variables, M):
    """Objective function to minimise additional thickness required to find a feasible thrust network

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    if isinstance(M, list):
        M = M[0]

    k = M.k
    nb = len(M.fixed)
    n = M.X.shape[0]

    check = k

    f = 0.0

    if 'xyb' in M.variables:
        check = check + 2*nb
    if 'zb' in M.variables:
        check = check + nb
    if 'tub' in M.variables:
        tub = variables[check: check + n]
        f += sum(tub**2)
        M.tub = tub
        check = check + n
    if 'tlb' in M.variables:
        tlb = variables[check: check + n]
        f += sum(tlb**2)
        M.tlb = tlb
        check = check + n
    if 'tlb' in M.variables:
        tlb = variables[check: check + n]
        f += sum(tlb**2)
        M.tlb = tlb
        check = check + n
    if 'tub_reac' in M.variables:
        tub_reac = variables[check: check + 2*nb]
        M.tub_reac = tub_reac
        f += sum(tub_reac[:nb]**2 + tub_reac[nb:]**2)  # this is not really 100% ok for 3D. x^2 +y^2 != (x + y)^2
        check = check + 2*nb

    return f


def f_constant(variables, M):
    """Constant or feasible objective function f=1

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    f = 1.0

    return f


def f_reduce_thk(variables, M):
    """Objective function to reduce the thickness of the structure

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.

    Note
    ----
        Please respect the order of the variables.
    """

    return variables[-1]


def f_tight_crosssection(variables, M):
    """Objective function to tight the cross section using normal vectors

    Parameters
    ----------
    variables : array (k x 1)
        Variables to pass to the function.
    M : :class:`~compas_tno.problems.Problem`
        The class with necessary matrices, or arguments, to compute the objective function

    Returns
    -------
    f : float
        The value of the objective function.
    """

    return -1 * variables[-1]



# from compas_tno.diagrams import FormDiagram
from compas.geometry import intersection_segment_segment_xy
from compas.geometry import distance_point_point_xy
from compas.geometry import Translation
from compas.geometry import Scale
from compas.geometry._core.distance import sort_points_xy
from compas.geometry._core.distance import closest_point_in_cloud
from compas.geometry import Point
from compas.geometry import Frame

from compas.datastructures import Mesh
from numpy import zeros
from random import shuffle


def is_point_in_cloud(point, cloud, tol=1e-6):
    if len(cloud) == 0:
        return False
    return closest_point_in_cloud(point, cloud)[0] < tol


def split_intersection_lines(lines, tol=1e-6):
    """Split lines accorting to their intersection

    Parameters
    ----------
    lines : [[list]]
        List of lines

    Returns
    -------
    clean_lines
        Lines split at the intersections
    """

    lines = [[[pt1[0], pt1[1], 0.0], [pt2[0], pt2[1], 0.0]] for pt1, pt2 in lines]
    clean_lines = []
    dict_lines = {i: [] for i in range(len(lines))}  # dict to store the inner points intersected

    # find intersections and store them as the inner points of given segments
    i = 0
    for line in lines:
        for line_ in lines:
            if line == line_:
                continue
            pt = intersection_segment_segment_xy(line, line_)
            if pt:
                if is_point_in_cloud(pt, line):
                    continue
                if not is_point_in_cloud(pt, dict_lines[i]):  # pt not in dict_lines[i]:
                    dict_lines[i].append(pt)
        i += 1

    # split lines containing inner intersections
    for key in dict_lines:
        line = lines[key]
        intpoints = dict_lines[key]
        if intpoints:
            np = len(intpoints)
            intsorted = sort_points_xy(line[0], dict_lines[key])
            startline = [line[0], intsorted[0][1]]
            endline = [intsorted[-1][1], line[1]]
            if distance_point_point_xy(*startline) > tol:
                clean_lines.append(startline)
            if distance_point_point_xy(*endline) > tol:
                clean_lines.append(endline)
            if np > 1:
                for k in range(np - 1):
                    intline = [intsorted[k][1], intsorted[k + 1][1]]
                    if distance_point_point_xy(*intline) > tol:
                        clean_lines.append(intline)
        else:
            if distance_point_point_xy(*line) > tol:
                clean_lines.append(line)

    return clean_lines


def form_add_lines_support(form, loaded_node, supports):
    """Add direct load paths from a chosen node to the chosen supports

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The FormDiagram to modify
    loaded_node : int
        Key of the node which will be the starting point for the additional lines
    supports : [int]
        List with the keys of the supports that should be linked to the loaded node

    Returns
    -------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram with the modifications

    new_loaded_node : int
        Key for the loaded node at the same position
    """

    from compas_tno.diagrams import FormDiagram

    text = {}
    new_lines = []
    text[loaded_node] = loaded_node

    xp, yp, _ = form.vertex_coordinates(loaded_node)
    fixed_coords = [form.vertex_coordinates(vertex) for vertex in form.vertices_where({'is_fixed': True})]
    parameters = form.parameters

    lines = form.to_lines()
    support_lines = []
    points = []

    for support in supports:
        xs, ys, _ = form.vertex_coordinates(support)
        support_line = [[xs, ys, 0.0], [xp, yp, 0.0]]
        support_lines.append(support_line)

    for line in lines:
        int_pt1 = intersection_segment_segment_xy(support_lines[0], line)
        int_pt2 = intersection_segment_segment_xy(support_lines[1], line)
        if int_pt1 or int_pt2:
            int_pt = int_pt1 or int_pt2
            points.append(int_pt)
            if int_pt == line[0] or int_pt == line[1]:
                new_lines.append(line)  # necessary?
                pass
            else:
                cutline1 = [line[0], int_pt]
                cutline2 = [int_pt, line[1]]
                new_lines.append(cutline1)
                new_lines.append(cutline2)
        else:
            new_lines.append(line)

    for i in range(len(points) - 1):
        j = i + 1
        if distance_point_point_xy(points[i], points[j]) > 1e-3:
            new_lines.append([points[i], points[j]])

    form = FormDiagram.from_lines(new_lines)

    # find new vertices
    for vertex in form.vertices():
        coord = form.vertex_coordinates(vertex)
        dist = distance_point_point_xy(coord, [xp, yp])
        if dist < 1e-3:
            new_loaded_node = vertex
            continue
        for coord_fix in fixed_coords:
            dist = distance_point_point_xy(coord, coord_fix)
            if dist < 1e-3:
                form.vertex_attribute(vertex, 'is_fixed', True)

    form.parameters = parameters

    return form, new_loaded_node


def store_inds(form, ind_edges=[]):
    """Store independent edges as an attribute.
    Note: The midpoint of the independent edges is stored.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to store
    ind_edges : list, optional
        List with the independent edges, by default [] in which edges

    Returns
    -------
    None
        Attribute included to the form diagram
    """

    points = []
    if len(ind_edges) == 0:
        for u, v in form.edges_where({'is_ind': True}):
            points.append(Point(*(form.edge_midpoint(u, v)[:2] + [0])))
    else:
        for u, v in ind_edges:
            points.append(Point(*(form.edge_midpoint(u, v)[:2] + [0])))
    form.attributes['indset'] = points

    return


def retrieve_inds(form, indset=[]):
    """Store independent edges as an attribute.
    Note: The midpoint of the independent edges is stored.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to store
    indset : list
        List with the independent edges

    Returns
    -------
    None
        Attribute included to the form diagram
    """

    print('WIP')

    # from compas.utilities import reverse_geometric_key

    # points = []

    # indset = [a if not isinstance(a, str) else reverse_geometric_key(a) for a in form.attributes['indset']]
    # for pt in indset:
    #     points.append(Point(*(form.edge_midpoint(u, v)[:2] + [0])))
    #     for u, v in form.edges_where({'_is_edge': True}):
    #         index = problem.uv_i[(u, v)]
    #         edgemid = Point(*(form.edge_midpoint(u, v)[:2] + [0]))
    #         for pt in indset:
    #             if distance_point_point_xy(edgemid, pt) < tol_old_ind:
    #                 ind.append(index)
    #                 break
    #         if index in ind:
    #             indset.remove(pt)
    # form.attributes['indset'] = points

    return


def retrieve_form_polylines(form, singularities=[]):
    """Store independent edges as an attribute.
    Note: The midpoint of the independent edges is stored.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram to store
    singularities : list
        List with index of singularities

    Returns
    -------
    polylines : list[list]
        List with the edges forming polylines
    """

    edges_visited = []
    edges = list(form.edges_where({'_is_edge': True}))

    for edge in edges:
        if edge in edges_visited:
            continue
        edge

    return


def mesh_remove_two_valent_nodes(mesh, cls=Mesh, delete_boundary_face=True):
    """Remove 2 valent nodes from a mesh or form diagram

    Parameters
    ----------
    mesh : :class:`~compas.datastructures.Mesh`
        Mesh to perform operations
    cls : class, optional
        Class to return the mesh, by default Mesh

    Returns
    -------
    mesh
        Mesh without nodes
    """

    uv_i = {(u, v): index for index, (u, v) in enumerate(mesh.edges())}
    delete_indices = []
    newedges = []
    for key in mesh.vertices():
        deg = mesh.vertex_degree(key)
        if deg == 2:
            pts = mesh.vertex_neighbors(key)
            ed = mesh.vertex_edges(key)
            delete_indices.append(uv_i[ed[0]])
            delete_indices.append(uv_i[ed[1]])
            newedges.append([mesh.vertex_coordinates(a) for a in pts])

    lines = mesh.to_lines()
    new_lines = []

    for i, line in enumerate(lines):
        if i in delete_indices:
            continue
        new_lines.append(line)

    for edge in newedges:
        new_lines.append(edge)

    mesh = cls.from_lines(new_lines, delete_boundary_face=delete_boundary_face)

    return mesh


def form_parabolic_slide(form, delta, y0=0.0, y1=10.0):
    """Modify the form diagram applying a parabolic displacement profile to the nodes

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The Form Diagram
    delta : float
        Maximum distance applied to the nodes
    y0 : float
        Start ordinate to apply the parabolic pattern
    y1 : float
        End ordinate to apply the parabolic pattern

    Returns
    -------
    form : FormDiagram
        The form diagram with the modifications

    """

    yc = ((y1 - y0)/2)
    for vertex in form.vertices():
        x, y, _ = form.vertex_coordinates(vertex)
        dy = min(y - y0, y1 - y)
        if abs(dy) > 1e-3:
            dx = delta * (1 - ((dy - yc)/yc)**2)
            form.vertex_attribute(vertex, 'x', x + dx)

    return form


def move_pattern_to_origin(mesh, corners=[[0.0, 0.0], [10.0, 0.0], [10.0, 10.0], [0.0, 10.0]], fix_corners=True):
    """Modify the form diagram to coincide the lower corner of its bounding box with the corners provided. Note: translation and scaling are involved.

    Parameters
    ----------
    mesh : :class:`~compas.datastructures.Mesh`
        The form diagram adopted
    corners : [pts]
        A list with four points for the corner
    fix_corners : bool, optional
        Whether or not the corners provided should be considered fixed

    Returns
    -------
    None
        Form is updated in place
    """

    xmax_, ymax_ = max([a[0] for a in corners]), max([a[1] for a in corners])
    xmin_, ymin_ = min([a[0] for a in corners]), min([a[1] for a in corners])

    trans = []
    factors = []

    bbox = mesh.bounding_box_xy()
    xmin, ymin = min([m[0] for m in bbox]), min([m[1] for m in bbox])
    # print('bbox beginning', bbox)

    if abs(xmin - xmin_) > 0.0 or abs(ymin - ymin_) > 0.0:
        dx, dy = -(xmin - xmin_), -(ymin - ymin_)
        trans = [dx, dy, 0.0]
        translation = Translation.from_vector(trans)
        mesh.transform(translation)

    bbox = mesh.bounding_box_xy()
    xmax, ymax = max([m[0] for m in bbox]), max([m[1] for m in bbox])
    # print('bbox 2', bbox)
    # print(xmax, ymax)

    if abs(ymax_ - ymax) > 0.0 or abs(xmax_ - xmax) > 0.0:
        factors = [xmax_/xmax, ymax_/ymax, 0.0]
        scale = Scale.from_factors(factors, frame=Frame(Point(xmin_, ymin_, 0.0), (1, 0, 0), (0, 1, 0)))
        mesh.transform(scale)

    if fix_corners:
        for key in mesh.vertices():
            pt = mesh.vertex_coordinates(key)
            for corner in corners:
                dist = distance_point_point_xy(pt, corner)
                if dist < 1e-3:
                    mesh.vertex_attribute(key, 'is_fixed', True)
                    break

    return


def slide_diagram(form, delta=0.5, y0=0.0, y1=10.0, tappered=False):
    """Apply a parabolic sliding to the nodes towards +x direction. Sliding profile is defined upon pattern's height (y-coordinate)

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    delta : float, optional
        Delta distance to apply to the nodes, by default 0.5
    y0 : float, optional
        Start of parabolic profile, by default 0.0
    y1 : float, optional
        End of parabolic profile, by default 10.0
    """

    yc = ((y1 - y0)/2)
    for vertex in form.vertices_where({'is_fixed': False}):
        x, y, _ = form.vertex_coordinates(vertex)
        dy = min(y - y0, y1 - y)
        if tappered:
            delta_i = delta * (x - y0)/(y1 - y0)
        else:
            delta_i = delta
        if abs(dy) > 1e-3:
            # dx = delta * (1 - ((dy - yc)/yc)**2)
            dx = delta_i * (1 - ((dy - yc)/yc)**2)
            form.vertex_attribute(vertex, 'x', x + dx)


def slide_pattern_inwards(form, delta=0.1, y0=0.0, y1=10.0, x0=0.0, x1=10.0, tol=0.01):
    """ Set parabolic vault heights.

    Parameters
    ----------

    Returns
    -------
    intrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the intrados of the shape
    extrados : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the extrados of the shape
    middle : :class:`~compas_tno.shapes.MeshDos`
        A MeshDos for the middle of the shape

    Notes
    ----------------------
    Position of the quadrants is as in the schema below:

        Q3
    Q2      Q1
        Q4

    """

    xc = (x0 + x1)/2
    yc = (y0 + y1)/2

    for vertex in form.vertices():
        if form.vertex_attribute(vertex, 'is_fixed'):
            continue
        x, y, _ = form.vertex_coordinates(vertex)

        quad = None
        dxi = 0.0
        dyi = 0.0

        if y <= y0 + (y1 - y0)/(x1 - x0) * (x - x0) - tol and y >= y1 - (y1 - y0)/(x1 - x0) * (x - x0) + tol:  # Q1
            quad = 'Q1'
        elif y >= y0 + (y1 - y0)/(x1 - x0) * (x - x0) + tol and y >= y1 - (y1 - y0)/(x1 - x0) * (x - x0) + tol:  # Q3
            quad = 'Q3'
        elif y >= y0 + (y1 - y0)/(x1 - x0) * (x - x0) + tol and y <= y1 - (y1 - y0)/(x1 - x0) * (x - x0) - tol:  # Q2
            quad = 'Q2'
        elif y <= y0 + (y1 - y0)/(x1 - x0) * (x - x0) - tol and y <= y1 - (y1 - y0)/(x1 - x0) * (x - x0) - tol:  # Q4
            quad = 'Q4'
        else:  # diagonals
            continue

        # factor = ((xc - dx)/xc)
        dy = min(y - y0, y1 - y)/yc  # 1 in center and 0 in corners (linear) - x
        dx = min(x - x0, x1 - x)/xc  # 1 in center and 0 in corners (linear) - y
        # x0_par = y0_par = x0 + (1 - dy/yc)
        # x1_par = y1_par = x1 - (1 - dy/yc)
        # dy_par = dy * (1 + (dx)/xc)
        # dx_par = dx * (1 + (dy)/yc)
        if quad in ['Q3', 'Q4']:
            dxa = (dx - dy) * 1/(1 - dy)
            mag_y = delta * (1 - ((1) - dxa)**2) * (1 - dy)
        if quad in ['Q1', 'Q2']:
            dya = (dy - dx) * 1/(1 - dx)
            mag_x = delta * (1 - ((1) - dya)**2) * (1 - dx)
        # mag_x = delta * (1 - ((dy - yc)/yc)**2) * ((xc - dx)/xc)
        # mag_y = delta * (1 - ((dx - xc)/xc)**2) * ((yc - dy)/yc)
        # mag_x = delta * (1 - ((1) - dya)**2) #* (1 - dx)
        # mag_y = delta * (1 - ((1) - dxa)**2) #* (1 - dy)

        if quad == 'Q1':
            dxi = - mag_x
        elif quad == 'Q3':
            dyi = - mag_y
        elif quad == 'Q2':  # Q2
            dxi = + mag_x
        elif quad == 'Q4':  # Q4
            dyi = mag_y
        else:  # diagonals
            continue

        # print(vertex, 'x, y;', x, y, 'mag xy:', mag_x, mag_y)

        form.vertex_attribute(vertex, 'x', x + dxi)
        form.vertex_attribute(vertex, 'y', y + dyi)


def displacement_map_parabola(form, y0=0.0, y1=10.0):
    """Create the displacement map for a 1D parabola sliding of the structural pattern.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    delta : float, optional
        Delta distance to apply to the nodes, by default 0.5
    y0 : float, optional
        Start of parabolic profile, by default 0.0
    y1 : float, optional
        End of parabolic profile, by default 10.0

    Returns
    -------
    dX: array [n, 2]
        Displacement field in plan.
    """

    n = form.number_of_vertices()
    k = - 4/(y1 - y0)**2
    dX = zeros((n, 2))

    for i, vertex in enumerate(form.vertices()):
        if form.vertex_attribute(vertex, 'is_fixed'):
            continue
        x, y, _ = form.vertex_coordinates(vertex)
        dyi = 0.0
        dxi = k * (y - y0) * (y - y1)
        dX[i] = [dxi, dyi]

    return dX


def displacement_map_4parabolas(form, y0=0.0, y1=10.0, x0=0.0, x1=10.0, tol=0.1):
    """Create the displacement map based on a paraboloid sliding the structural pattern.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    delta : float, optional
        Delta distance to apply to the nodes, by default 0.5
    y0 : float, optional
        Start of parabolic profile, by default 0.0
    y1 : float, optional
        End of parabolic profile, by default 10.0

    Returns
    -------
    dX: array [n, 2]
        Displacement field in plan.
    """

    n = form.number_of_vertices()
    kx = - 4/(y1 - y0) ** 2
    ky = - 4/(x1 - x0) ** 2
    dX = zeros((n, 2))

    for i, vertex in enumerate(form.vertices()):
        if form.vertex_attribute(vertex, 'is_fixed'):
            continue
        x, y, _ = form.vertex_coordinates(vertex)
        dxi = 0.0
        dyi = 0.0

        if y <= y0 + (y1 - y0)/(x1 - x0) * (x - x0) - tol and y >= y1 - (y1 - y0)/(x1 - x0) * (x - x0) + tol:  # Q1
            dxi = - kx * (y - y0) * (y - y1)
        elif y >= y0 + (y1 - y0)/(x1 - x0) * (x - x0) + tol and y >= y1 - (y1 - y0)/(x1 - x0) * (x - x0) + tol:  # Q3
            dyi = - ky * (x - x0) * (x - x1)
        elif y >= y0 + (y1 - y0)/(x1 - x0) * (x - x0) + tol and y <= y1 - (y1 - y0)/(x1 - x0) * (x - x0) - tol:  # Q2
            dxi = + kx * (y - y0) * (y - y1)
        elif y <= y0 + (y1 - y0)/(x1 - x0) * (x - x0) - tol and y <= y1 - (y1 - y0)/(x1 - x0) * (x - x0) - tol:  # Q4
            dyi = ky * (x - x0) * (x - x1)
        else:  # diagonals
            pass

        dX[i] = [dxi, dyi]

    return dX


def displacement_map_paraboloid(form, xc=5.0, yc=5.0, radius=5.0):
    """Create the displacement map based on a paraboloid sliding the structural pattern.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem
    delta : float, optional
        Delta distance to apply to the nodes, by default 0.5
    y0 : float, optional
        Start of parabolic profile, by default 0.0
    y1 : float, optional
        End of parabolic profile, by default 10.0

    Returns
    -------
    dX: array [n, 2]
        Displacement field in plan.
    """

    n = form.number_of_vertices()
    k = 1/radius**2
    dX = zeros((n, 2))

    for i, vertex in enumerate(form.vertices()):
        if form.vertex_attribute(vertex, 'is_fixed'):
            continue
        x, y, _ = form.vertex_coordinates(vertex)
        dxi = k * (x - xc) ** 2 * 2 * (xc - x) / radius
        dyi = k * (y - yc) ** 2 * 2 * (yc - y) / radius

        dX[i] = [dxi, dyi]

    return dX


def shuffle_diagram(form):
    """Shuffle the form diagram reordering the edges. It keeps the support points and '_is_edge' attributes.

    Parameters
    ----------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The form diagram of the problem

    Returns
    -------
    form : :class:`~compas_tno.diagrams.FormDiagram`
        The shuffled diagram
    """

    from compas_tno.diagrams import FormDiagram
    tol = 1e-6

    lines = form.to_lines()
    supports = [form.vertex_coordinates(key) for key in form.vertices_where({'is_fixed': True})]
    no_edges = [form.edge_midpoint(*edge) for edge in form.edges_where({'_is_edge': False})]

    shuffle(lines)
    form = FormDiagram.from_lines(lines)

    for key in form.vertices():
        pt = form.vertex_coordinates(key)
        for support in supports:
            dist = distance_point_point_xy(pt, support)
            if dist < tol:
                form.vertex_attribute(key, 'is_fixed', True)
                break

    for edge in form.edges():
        pt = form.edge_midpoint(*edge)
        for midpoint in no_edges:
            dist = distance_point_point_xy(pt, midpoint)
            if dist < tol:
                form.edge_attribute(edge, '_is_edge', False)
                break

    return form



import json


__all__ = [
    'update_json'
]


def update_json(infile):
    """Update JSON for new compas edge data information

    Parameters
    ----------
    infile : str
        The ``.json`` address of the file in the old format

    Returns
    -------
    outfile
        The address of the new, saved file.
    """

    outfile = ''.join(infile.split('.')[:-1]) + '_review.' + infile.split('.')[-1]

    file = open(infile)
    data = json.load(file)

    edge_data = {}
    for key in data['data']['edgedata']:
        a, b = key.split('-')
        newkey = str((int(a), int(b)))
        edge_data[newkey] = data['data']['edgedata'][key]

    data['data']['edgedata'] = edge_data

    with open(outfile, 'w') as savefile:
        json.dump(data, savefile)

    return outfile



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino

from functools import partial
from math import sqrt
from math import pi
from compas_tno.rhino.diagramartist import DiagramArtist
from compas.utilities import color_to_colordict
from compas.geometry import add_vectors
from compas.geometry import scale_vector
from compas.geometry import length_vector
from compas.geometry import norm_vector

from compas.utilities import is_color_rgb

colordict = partial(color_to_colordict, colorformat='rgb', normalize=False)


__all__ = ['FormArtist']


class FormArtist(DiagramArtist):
    """Artist for form diagram in TNO.

    Parameters
    ----------
    form: compas_tno.diagrams.FormDiagram
        The form diagram to draw.

    Attributes
    ----------
    mesh : :class:`compas_tno.diagrams.FormDiagram`
        The form diagram associated with the artist.
    color_compression : 3-tuple
        Default color for compression.
    color_tension : 3-tuple
        Default color for tension.
    scale_forces : float
        Scale factor for the force pipes.
    tol_forces : float
        Tolerance for force magnitudes.
    vertices : list
        The vertices to include in the drawing.
        Default is all vertices.
    edges : list
        The edges to include in the drawing.
        Default is all edges.
    faces : list
        The faces to include in the drawing.
        Default is all faces.
    vertex_xyz : dict
        The view coordinates of the vertices.
        Default is to use the actual vertex coordinates.
    vertex_color : dict
        Mapping between vertices and colors.
        Default is to use the default color for vertices.
    edge_color : dict
        Mapping between edges and colors.
        Default is to use the default color for edges.
    face_color : dict
        Mapping between faces and colors.
        Default is to use the default color for faces.
    vertex_text : dict
        Mapping between vertices and text labels.
    edge_text : dict
        Mapping between edges and text labels.
    face_text : dict
        Mapping between faces and text labels.
    """

    def __init__(self, form, layer=None):
        super(FormArtist, self).__init__(form, layer=layer)
        self.color_compression = (255, 0, 0)
        self.color_tension = (0, 255, 0)
        self.color_reaction = (125, 125, 125)
        self.color_mesh_thrust = (0, 0, 0)
        self.color_mesh_intrados = (125, 125, 125)
        self.color_mesh_extrados = (125, 125, 125)
        self.color_mesh_middle = (0, 0, 0)
        self.color_vertex_extrados = (0, 125, 0)
        self.color_vertex_intrados = (0, 0, 255)
        self.color_faces = (0, 0, 0)
        self.scale_forces = 0.001
        self.pipes_scale = 0.01
        self.scale_line = 0.1
        self.tol_forces = 0.001
        self.radius_sphere = 0.15
        self.layer = 'FormDiagram'

        self._mesh = None
        self._vertices = None
        self._edges = None
        self._faces = None
        self._color = None
        self._vertex_xyz = None
        self._vertex_color = None
        self._vertex_text = None
        self._vertex_size = None
        self._edge_color = None
        self._edge_text = None
        self._edge_width = None
        self._face_color = None
        self._face_text = None

        self._vertexcollection = None
        self._edgecollection = None
        self._facecollection = None
        self._vertexnormalcollection = None
        self._facenormalcollection = None
        self._vertexlabelcollection = None
        self._edgelabelcollection = None
        self._facelabelcollection = None

        self.mesh = form

    @property
    def mesh(self):
        return self._mesh

    @mesh.setter
    def mesh(self, mesh):
        self._mesh = mesh
        self._vertex_xyz = None

    @property
    def vertices(self):
        if self._vertices is None:
            self._vertices = list(self.mesh.vertices())
        return self._vertices

    @vertices.setter
    def vertices(self, vertices):
        self._vertices = vertices

    @property
    def edges(self):
        if self._edges is None:
            self._edges = list(self.mesh.edges())
        return self._edges

    @edges.setter
    def edges(self, edges):
        self._edges = edges

    @property
    def faces(self):
        if self._faces is None:
            self._faces = list(self.mesh.faces())
        return self._faces

    @faces.setter
    def faces(self, faces):
        self._faces = faces

    @property
    def color(self):
        if not self._color:
            self._color = self.default_color
        return self._color

    @color.setter
    def color(self, color):
        if is_color_rgb(color):
            self._color = color

    @property
    def vertex_xyz(self):
        if self._vertex_xyz is None:
            return {vertex: self.mesh.vertex_attributes(vertex, 'xyz') for vertex in self.mesh.vertices()}
        return self._vertex_xyz

    @vertex_xyz.setter
    def vertex_xyz(self, vertex_xyz):
        self._vertex_xyz = vertex_xyz

    @property
    def vertex_color(self):
        if self._vertex_color is None:
            self._vertex_color = {vertex: self.default_vertexcolor for vertex in self.mesh.vertices()}
        return self._vertex_color

    @vertex_color.setter
    def vertex_color(self, vertex_color):
        if isinstance(vertex_color, dict):
            self._vertex_color = vertex_color
        elif is_color_rgb(vertex_color):
            self._vertex_color = {vertex: vertex_color for vertex in self.mesh.vertices()}

    @property
    def vertex_text(self):
        if self._vertex_text is None:
            self._vertex_text = {vertex: str(vertex) for vertex in self.mesh.vertices()}
        return self._vertex_text

    @vertex_text.setter
    def vertex_text(self, text):
        if text == 'key':
            self._vertex_text = {vertex: str(vertex) for vertex in self.mesh.vertices()}
        elif text == 'index':
            self._vertex_text = {vertex: str(index) for index, vertex in enumerate(self.mesh.vertices())}
        elif isinstance(text, dict):
            self._vertex_text = text

    @property
    def vertex_size(self):
        if not self._vertex_size:
            self._vertex_size = {vertex: self.default_vertexsize for vertex in self.mesh.vertices()}
        return self._vertex_size

    @vertex_size.setter
    def vertex_size(self, vertexsize):
        if isinstance(vertexsize, dict):
            self._vertex_size = vertexsize
        elif isinstance(vertexsize, (int, float)):
            self._vertex_size = {vertex: vertexsize for vertex in self.mesh.vertices()}

    @property
    def edge_color(self):
        if self._edge_color is None:
            self._edge_color = {edge: self.default_edgecolor for edge in self.mesh.edges()}
        return self._edge_color

    @edge_color.setter
    def edge_color(self, edge_color):
        if isinstance(edge_color, dict):
            self._edge_color = edge_color
        elif is_color_rgb(edge_color):
            self._edge_color = {edge: edge_color for edge in self.mesh.edges()}

    @property
    def edge_text(self):
        if self._edge_text is None:
            self._edge_text = {edge: "{}-{}".format(*edge) for edge in self.mesh.edges()}
        return self._edge_text

    @edge_text.setter
    def edge_text(self, text):
        if text == 'key':
            self._edge_text = {edge: "{}-{}".format(*edge) for edge in self.mesh.edges()}
        elif text == 'index':
            self._edge_text = {edge: str(index) for index, edge in enumerate(self.mesh.edges())}
        elif isinstance(text, dict):
            self._edge_text = text

    @property
    def edge_width(self):
        if not self._edge_width:
            self._edge_width = {edge: self.default_edgewidth for edge in self.mesh.edges()}
        return self._edge_width

    @edge_width.setter
    def edge_width(self, edgewidth):
        if isinstance(edgewidth, dict):
            self._edge_width = edgewidth
        elif isinstance(edgewidth, (int, float)):
            self._edge_width = {edge: edgewidth for edge in self.mesh.edges()}

    @property
    def face_color(self):
        if self._face_color is None:
            self._face_color = {face: self.default_facecolor for face in self.mesh.faces()}
        return self._face_color

    @face_color.setter
    def face_color(self, face_color):
        if isinstance(face_color, dict):
            self._face_color = face_color
        elif is_color_rgb(face_color):
            self._face_color = {face: face_color for face in self.mesh.faces()}

    @property
    def face_text(self):
        if self._face_text is None:
            self._face_text = {face: str(face) for face in self.mesh.faces()}
        return self._face_text

    @face_text.setter
    def face_text(self, text):
        if text == 'key':
            self._face_text = {face: str(face) for face in self.mesh.faces()}
        elif text == 'index':
            self._face_text = {face: str(index) for index, face in enumerate(self.mesh.faces())}
        elif isinstance(text, dict):
            self._face_text = text

    # def draw_edges(self, edges=None, color=None, displacement=None, layer='Thrust'):
    #     """Draw a selection of edges.

    #     Parameters
    #     ----------
    #     edges : list, optional
    #         A selection of edges to draw.
    #         The default is ``None``, in which case all edges are drawn.
    #     color : tuple or dict of tuple, optional
    #         The color specififcation for the edges.
    #         The default color is black, ``(0, 0, 0)``.

    #     Returns
    #     -------
    #     list
    #         The GUIDs of the created Rhino objects.

    #     """
    #     edges = edges or list(self.diagram.edges_where({'_is_edge': True}))
    #     vertex_xyz = self.vertex_xyz
    #     if displacement:
    #         for key in vertex_xyz:
    #             vertex_xyz[key][0] += displacement[0]
    #             vertex_xyz[key][1] += displacement[1]
    #     edge_color = colordict(color, edges, default=self.color_edges)
    #     lines = []
    #     for edge in edges:
    #         lines.append({
    #             'start': vertex_xyz[edge[0]],
    #             'end': vertex_xyz[edge[1]],
    #             'color': edge_color[edge],
    #             'name': "{}.edge.{}-{}".format(self.diagram.name, *edge)})
    #     return compas_rhino.draw_lines(lines, layer=layer, clear=False, redraw=False)

    def redraw(self):

        compas_rhino.rs.EnableRedraw(True)

        return

    def draw_mesh(self, displacement=None):
        """Draw a mesh for the thrust network.

        Parameters
        ----------
        edges : list, optional
            A selection of edges to draw.
            The default is ``None``, in which case all edges are drawn.
        displacement : list, optional
            A displacement to add mesh to the scene.

        Returns
        -------
        list
            The GUIDs of the created Rhino object.

        """
        vertices, faces = self.diagram.to_vertices_and_faces()
        return compas_rhino.draw_mesh(vertices, faces, name="Thrust", color=self.color_mesh_thrust, disjoint=True, layer="Thrust-Mesh")

    def draw_thrust(self, scale_width=None, layer="FormDiagram::Thrust", tol=1e-2):
        """Draw the thrust network as a set of lines with width proportional to the square of the carried force.

        Parameters
        ----------
        edges : list, optional
            A selection of edges to draw.
            The default is ``None``, in which case all edges are drawn.
        displacement : list, optional
            A displacement to add mesh to the scene.

        Returns
        -------
        list
            The GUIDs of the created Rhino object.

        """

        layer = layer or self.layer
        color = self.color_compression
        scale_width = scale_width or self.scale_line
        lines = []

        for u, v in self.diagram.edges_where({'_is_edge': True}):
            Xu = self.diagram.vertex_coordinates(u)
            Xv = self.diagram.vertex_coordinates(v)

            q = self.diagram.edge_attribute((u, v), 'q')
            length = self.diagram.edge_length(u, v)
            force = abs(q*length)

            if force < tol:
                continue

            radius = sqrt(force)

            thk = radius * scale_width

            lines.append({'start': Xu,
                          'end': Xv,
                          'color': color,
                          'width': thk
                          }
                         )

        return compas_rhino.draw_lines(lines, layer=layer, clear=False, redraw=True)

    def draw_cracks(self, color_intrados=(0, 0, 200), color_extrados=(0, 200, 0), layer=None, tol=10e-5):
        """Draw the intersection of the thrust network with intrados and extrados.

        Parameters
        ----------
        vertices : list, optional
            A selection of vertices to draw.
            The default is ``None``, in which case all vertices are drawn.
        displacement : list, optional
            A displacement to add mesh to the scene.
        tol : float, optional
            Acceptable error to find an intersection.

        Returns
        -------
        tuple with 2 lists
            List of the GUIDs of the created Rhino object, and the keys associated to it in the form diagram.

        """
        layer = layer or self.layer
        form = self.diagram
        intra_vertices = []
        extra_vertices = []
        intra_keys = []
        extra_keys = []
        for key in form.vertices():
            x, y, z = form.vertex_coordinates(key)
            lb = form.vertex_attribute(key, 'lb')
            ub = form.vertex_attribute(key, 'ub')
            if lb:
                if abs(z - lb) < tol:
                    intra_keys.append(key)
                    intra_vertices.append({
                        'pos': [x, y, z],
                        'name': "{}.vertex.{}".format("Intrados", key),
                        'color': self.color_vertex_intrados})
            if ub:
                if abs(z - ub) < tol:
                    extra_keys.append(key)
                    extra_vertices.append({
                        'pos': [x, y, z],
                        'name': "{}.vertex.{}".format("Extrados", key),
                        'color': self.color_vertex_extrados})

        guids_intra = compas_rhino.draw_points(intra_vertices, layer=layer + '::Intrados', color=color_intrados, clear=False, redraw=False)
        guids_extra = compas_rhino.draw_points(extra_vertices, layer=layer + '::Extrados', color=color_extrados, clear=False, redraw=False)
        return (guids_intra + guids_extra, intra_keys + extra_keys)

    # update this function
    def draw_from_attributes(self, attribute='target', name='mesh', displacement=None, color=None, join_faces=True, layer=None):
        """Draw a copy mesh of the form diagram whose height is defined based on the a given attribure.

        Parameters
        ----------
        attribute : str
            Attribute to base the heights of the network on.
        scale : float, optional
            The scaling factor for the load force vectors.
            The default value is ``0.01``
        layer : str, optional
            The layer to draw the output.

        Returns
        -------
        list
            A list with the guid of the corresponding load force vectors in Rhino.

        Notes
        -----
        The magnitude of the externally applied load at a vetex the attribute  `pz`.

        """

        faces = list(self.diagram.faces())
        layer = layer or self.layer
        vertex_xyz = self.vertex_xyz
        for key in vertex_xyz:
            z = self.diagram.vertex_attribute(key, attribute)  # Check my forms and remove this
            if type(z) == list:
                z = z[0]
            vertex_xyz[key][2] = z
            if displacement:
                vertex_xyz[key][0] += displacement[0]
                vertex_xyz[key][1] += displacement[1]
        face_color = colordict(color, faces, default=self.color_faces)
        facets = []
        for face in faces:
            facets.append({
                'points': [vertex_xyz[vertex] for vertex in self.diagram.face_vertices(face)],
                'name': "{}.face.{}".format(name, face),
                'color': face_color[face]})
        guids = compas_rhino.draw_faces(facets, layer=layer, clear=False, redraw=False)
        if not join_faces:
            return guids
        guid = compas_rhino.rs.JoinMeshes(guids, delete_input=True)
        compas_rhino.rs.ObjectLayer(guid, layer)
        compas_rhino.rs.ObjectName(guid, '{}'.format(name))
        if color:
            compas_rhino.rs.ObjectColor(guid, color)
        return

    def draw_loads(self, color=(255, 0, 0), scale=0.01, layer="FormDiagram::Loads", tol=1e-3):
        """Draw the externally applied loads at all vertices of the diagram.

        Parameters
        ----------
        color : list or tuple, optional
            The RGB color specification for load forces.
            The specification must be in integer format, with each component between 0 and 255.
        scale : float, optional
            The scaling factor for the load force vectors.
            The default value is ``0.01``
        layer : str, optional
            The layer to draw the output.

        Returns
        -------
        list
            A list with the guid of the corresponding load force vectors in Rhino.

        Notes
        -----
        The magnitude of the externally applied load at a vetex the attribute  `pz`.

        """
        vertex_xyz = self.vertex_xyz
        layer = layer or self.layer
        lines = []

        for vertex in self.diagram.vertices():
            a = vertex_xyz[vertex]
            pz = -1 * self.diagram.vertex_attribute(vertex, 'pz')
            load = scale_vector((0, 0, 1), scale * pz)
            b = add_vectors(a, load)
            lines.append({'start': a, 'end': b, 'color': color, 'arrow': "start"})

        return compas_rhino.draw_lines(lines, layer=layer, clear=False, redraw=False)

    def draw_reactions(self, color=None, scale=0.01, draw_as_pipes=False, layer="FormDiagram::Reactions", tol=1e-3):
        """Draw the reaction forces.

        Parameters
        ----------
        color : list or tuple
            The RGB color specification for load forces.
            The specification must be in integer format, with each component between 0 and 255.
        scale : float, optional
            The scaling factor for the load force vectors.
            The default value is ``0.01``
        layer : str, optional
            The layer to draw the output.

        Returns
        -------
        list
            A list with the guid of the corresponding load force vectors in Rhino.

        Notes
        -----
        The magnitude of the externally applied load at a vetex the attribute  `pz`.

        """
        layer = layer or self.layer
        color = color or self.color_reaction
        vertex_xyz = self.vertex_xyz
        lines = []
        cylinders = []
        for key in self.diagram.vertices_where({'is_fixed': True}):
            a = vertex_xyz[key]
            r = self.diagram.vertex_attributes(key, ['_rx', '_ry', '_rz'])
            if not any(r):  # If receives null vector or None
                continue
            r = scale_vector(r, -scale)
            if length_vector(r) < tol:
                continue

            b = add_vectors(a, r)
            lines.append({'start': a,
                          'end': b,
                          'color': color,
                          'arrow': "start"  # 'arrow': "end"
                          }
                         )
            if draw_as_pipes:
                force = self.pipes_scale * norm_vector(self.diagram.vertex_attributes(key, ['_rx', '_ry', '_rz']))
                print(force)
                cylinders.append({
                    'start': a,
                    'end': b,
                    'radius': sqrt(abs(force)/pi),
                    'color': color
                })

        if draw_as_pipes:
            return compas_rhino.draw_cylinders(cylinders, layer=layer, clear=False, redraw=False)
        else:
            return compas_rhino.draw_lines(lines, layer=layer, clear=False, redraw=False)

    def draw_forcepipes(self, color_compression=(255, 0, 0), color_tension=(0, 0, 255), compression_negative=True, tol=1e-3, layer=None):
        """Draw the forces in the internal edges as pipes with color and thickness matching the force value.

        Parameters
        ----------
        color_compression
        color_tension
        scale
        tol

        Returns
        -------
        list
            The GUIDs of the created Rhino objects.
        """
        layer = layer or self.layer
        vertex_xyz = self.vertex_xyz
        scale = self.pipes_scale
        cylinders = []
        for edge in self.diagram.edges_where({'_is_edge': True}):
            u, v = edge
            start = vertex_xyz[u]
            end = vertex_xyz[v]
            length = self.diagram.edge_length(*edge)
            q = self.diagram.edge_attribute(edge, 'q')
            force = q * length
            force = scale * force
            if abs(force) < tol:
                continue
            radius = sqrt(abs(force)/pi)
            if compression_negative:
                pipe_color = color_compression if force < 0 else color_tension
            else:
                pipe_color = color_tension if force < 0 else color_compression
            cylinders.append({
                'start': start,
                'end': end,
                'radius': radius,
                'color': pipe_color
            })
        return compas_rhino.draw_cylinders(cylinders, layer=layer, clear=False, redraw=False)
